"""Click DDS options used multiple times in __main__.py."""

# Imports
import click
from dds_cli.utils import multiple_help_text

# Args used multiple times
def email_arg(required, email="email", metavar="[EMAIL]", nargs=1):
    """
    Email positional argument standard definition.

    Use as decorator for commands.
    """
    return click.argument(email, metavar=metavar, nargs=nargs, type=str, required=required)


# Options used multiple times
def email_option(help_message, long="--email", short="-e", name="email", required=True):
    """
    Email option standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        short,
        name,
        required=required,
        type=str,
        help=help_message,
    )


def folder_option(
    help_message, long="--folder", short="-f", name="folder", required=False, multiple=False
):
    """
    Folder option standard definition.

    Use as decorator for commands.
    """
    if multiple:
        help_message += multiple_help_text(item="folder")
    return click.option(
        long,
        short,
        name,
        required=required,
        type=str,
        multiple=multiple,
        help=help_message,
    )


def num_threads_option(
    long="--num-threads",
    short="-nt",
    name="num_threads",
    required=False,
    default=4,
    show_default=True,
    help_message="Number of parallel threads to perform the delivery",
):
    """
    Num threads option standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        short,
        name,
        required=required,
        default=default,
        show_default=show_default,
        type=click.IntRange(1, 32),
        help=help_message,
    )


def project_option(
    required, long="--project", short="-p", name="project", help_message="Project ID."
):
    """
    Project option standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        short,
        name,
        required=required,
        type=str,
        help=help_message,
    )


def sort_projects_option(
    long="--sort",
    name="sort",
    choices=["id", "title", "pi", "status", "updated", "size", "usage", "cost"],
    case_sensitive=False,
    default="Updated",
    required=False,
    help_message="Which column to sort the project list by.",
):
    """
    Sort option standard definition.

    Use as decorator for commands where you wish to sort the projects.
    """
    return click.option(
        long,
        name,
        type=click.Choice(
            choices=choices,
            case_sensitive=case_sensitive,
        ),
        default=default,
        required=required,
        help=help_message,
    )


def source_option(
    help_message, option_type, long="--source", short="-s", name="source", required=False
):
    """
    Source option standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        short,
        name,
        required=required,
        type=option_type,
        multiple=True,
        help=help_message + multiple_help_text(item="source"),
    )


def source_path_file_option(
    long="--source-path-file",
    short="-spf",
    name="source_path_file",
    required=False,
    help_message="File containing path to files or directories.",
):
    """
    Source path file option standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        short,
        name,
        required=required,
        type=click.Path(exists=True),
        multiple=False,
        help=help_message,
    )


def username_option(
    long="--username",
    short="-u",
    name="username",
    required=False,
    help_message="Your Data Delivery System username.",
):
    """
    Username option standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        short,
        name,
        required=required,
        type=str,
        help=help_message,
    )


# Flags
def break_on_fail_flag(
    help_message,
    long="--break-on-fail",
    name="break_on_fail",
    show_default=True,
):
    """
    Break on fail flag standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        name,
        is_flag=True,
        default=False,
        show_default=show_default,
        help=help_message,
    )


def json_flag(help_message, long="--json", name="json", show_default=True):
    """
    Json flag standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        name,
        is_flag=True,
        default=False,
        show_default=show_default,
        help=help_message,
    )


def nomail_flag(
    help_message,
    long="--no-mail",
    name="no_mail",
    show_default=False,
):
    """
    No-email flag standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        name,
        is_flag=True,
        default=False,
        show_default=show_default,
        help=help_message,
    )


def silent_flag(
    help_message,
    long="--silent",
    name="silent",
    show_default=True,
):
    """
    Silent flag standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        name,
        is_flag=True,
        default=False,
        show_default=show_default,
        help=help_message,
    )


def size_flag(help_message, long="--size", name="size", show_default=True):
    """
    Size flag standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        name,
        is_flag=True,
        default=False,
        show_default=show_default,
        help=help_message,
    )


def tree_flag(help_message, long="--tree", name="tree", show_default=True):
    """
    Tree flag standard definition.

    Use as decorator for commands.
    """
    return click.option(
        long,
        name,
        is_flag=True,
        default=False,
        show_default=show_default,
        help=help_message,
    )


def usage_flag(help_message, long="--usage", name="usage", show_default=True):
    """
    Usage flag standard definition.

    Use as decorator for commands where displaying the usage may be desired.
    """
    return click.option(
        long, name, is_flag=True, default=False, show_default=show_default, help=help_message
    )


def users_flag(
    help_message,
    long="--users",
    name="users",
    show_default=True,
):
    """
    Users flag standard definition.

    Use as a decorator for commands.
    """
    return click.option(
        long,
        name,
        is_flag=True,
        default=False,
        show_default=show_default,
        help=help_message,
    )



""""Compressor module. Handles the compression of files."""

###############################################################################
# IMPORTS ########################################################### IMPORTS #
###############################################################################

# Standard library
import dataclasses
import logging
import os
import pathlib
import traceback


# Installed
import zstandard as zstd

# Own modules
from dds_cli import FileSegment

###############################################################################
# START LOGGING CONFIG ################################# START LOGGING CONFIG #
###############################################################################

LOG = logging.getLogger(__name__)

###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


class CompressionMagic:
    """Compression format signatures"""

    BZIP2 = b"BZh"
    LZIP = b"LZIP"
    RAR4 = b"Rar!\x1a\x07\x00"
    RAR5 = b"Rar!\x1a\x07\x01\x00"
    GZIP = b"\x1F\x8B"
    ZSTANDARD = b"(\xb5/\xfd"


@dataclasses.dataclass
class Compressor:
    """Handles operations relating to file compression."""

    algorithm: str = "zstandard"
    fmt_magic: dict = dataclasses.field(init=False)
    max_magic_len: int = dataclasses.field(init=False)

    def __post_init__(self):
        self.fmt_magic = {
            b"\x913HF": "hap",
            b"`\xea": "arj",
            b"_'\xa8\x89": "jar",
            b"ZOO ": "zoo",
            b"PK\x03\x04": "zip",
            b"\x1F\x8B": "gzip",
            b"UFA\xc6\xd2\xc1": "ufa",
            b"StuffIt ": "sit",
            b"Rar!\x1a\x07\x00": "rar v4.x",
            b"Rar!\x1a\x07\x01\x00": "rar v5",
            b"MAr0\x00": "mar",
            b"DMS!": "dms",
            b"CRUSH v": "cru",
            b"BZh": "bz2",
            b"-lh": "lha",
            b"(This fi": "hqx",
            b"!\x12": "ain",
            b"\x1a\x0b": "pak",
            b"(\xb5/\xfd": "zst",
        }
        self.max_magic_len = max(len(x) for x in self.fmt_magic)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
            return False  # uncomment to pass exception through

        return True

    # Static methods ###################### Static methods #
    @staticmethod
    def compress_file(
        file: pathlib.Path,
        chunk_size: int = FileSegment.SEGMENT_SIZE_RAW,
    ) -> (bytes):
        """Compresses file by reading it chunk by chunk."""

        LOG.debug("Started compression...")

        try:
            with file.open(mode="rb") as infile:

                # Initiate a Zstandard compressor
                cctzx = zstd.ZstdCompressor(write_checksum=True, level=4)

                # total_read = 0.0
                # Compress file chunk by chunk while reading
                with cctzx.stream_reader(infile) as compressor:
                    # while True:
                    #     chunk = compressor.read(chunk_size)
                    #     LOG.debug(type(chunk))
                    #     if not chunk:
                    #         break
                    #     yield
                    for chunk in iter(lambda: compressor.read(chunk_size), b""):
                        yield chunk
        except Exception as err:
            LOG.warning(str(err))

        LOG.debug("Compression finished.")

    @staticmethod
    def decompress_filechunks(chunks, outfile: pathlib.Path, **_):
        """Decompress file chunks"""

        saved, message = (False, "")

        # Decompressing file and saving
        LOG.debug("Decompressing...")
        try:
            with outfile.open(mode="wb+") as file:
                dctx = zstd.ZstdDecompressor()
                with dctx.stream_writer(file) as decompressor:
                    for chunk in chunks:
                        decompressor.write(chunk)

        except OSError as err:
            message = str(err)
            LOG.exception(message)
        else:
            saved = True
            LOG.debug("Decompression done.")

        return saved, message

    # Public methods ###################### Public methods #
    def is_compressed(self, file):
        """Checks if a file is compressed or not."""

        compressed, error = (False, "")
        try:
            with file.open(mode="rb") as f:
                file_start = f.read(self.max_magic_len)
                if file_start.startswith(tuple(x for x in self.fmt_magic)):
                    compressed = True
        except OSError as err:
            error = str(err)

        return compressed, error



"""S3 Connector module."""

###############################################################################
# IMPORTS ########################################################### IMPORTS #
###############################################################################

# Standard library
import dataclasses
import logging
import traceback
import requests

# Installed
import boto3
import botocore

# Own modules
from dds_cli import DDSEndpoint
from dds_cli import utils

###############################################################################
# LOGGING ########################################################### LOGGING #
###############################################################################

LOG = logging.getLogger(__name__)

###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


@dataclasses.dataclass
class S3Connector:
    """Connect to Simple Storage Service."""

    project_id: dataclasses.InitVar[str]
    token: dataclasses.InitVar[dict]
    safespring_project: str = dataclasses.field(init=False)
    keys: dict = dataclasses.field(init=False)
    url: str = dataclasses.field(init=False)
    bucketname: str = dataclasses.field(init=False)
    resource = None

    def __post_init__(self, project_id, token):
        """Initiate S3Connector object by getting s3 info from API."""
        (
            self.safespring_project,
            self.keys,
            self.url,
            self.bucketname,
        ) = self.__get_s3_info(project_id=project_id, token=token)

    # @connect_cloud
    def __enter__(self):
        """Enter context."""
        self.resource = self.connect()

        return self

    def __exit__(self, exc_type, exc_value, tb):
        """Close context manager, incl. connection."""
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
            return False  # uncomment to pass exception through

        return True

    def connect(self):
        """Connect to S3 resource."""
        # Connect to service
        try:
            session = boto3.session.Session()

            resource = session.resource(
                service_name="s3",
                endpoint_url=self.url,
                aws_access_key_id=self.keys["access_key"],
                aws_secret_access_key=self.keys["secret_key"],
            )
        except (boto3.exceptions.Boto3Error, botocore.exceptions.BotoCoreError) as err:
            LOG.warning(f"S3 connection failed: {err}")
            raise

        LOG.debug(f"Resource :{self.resource}")
        return resource

    # Static methods ############ Static methods #
    @staticmethod
    def __get_s3_info(project_id, token):
        """Get information required to connect to cloud."""
        # Perform request to API
        try:
            response = requests.get(
                DDSEndpoint.S3KEYS,
                params={"project": project_id},
                headers=token,
                timeout=DDSEndpoint.TIMEOUT,
            )
        except requests.exceptions.RequestException as err:
            LOG.warning(err)
            raise SystemExit from err

        # Error
        if not response.ok:
            message = f"Failed retrieving Safespring project name. Server response: {response.text}"
            LOG.warning(message)
            raise SystemExit(message)  # TODO: Change

        # Get s3 info
        s3info = utils.get_json_response(response=response)

        safespring_project, keys, url, bucket = (
            s3info.get("safespring_project"),
            s3info.get("keys"),
            s3info.get("url"),
            s3info.get("bucket"),
        )
        if None in [safespring_project, keys, url, bucket]:
            raise SystemExit("Missing safespring information in response.")  # TODO: change

        return safespring_project, keys, url, bucket



"""Admin class, adds ."""

####################################################################################################
# IMPORTS ################################################################################ IMPORTS #
###################################################################################################

# Standard library
import logging

# Installed
import http
import requests
import simplejson


# Own modules
import dds_cli
import dds_cli.auth
import dds_cli.base
import dds_cli.exceptions


####################################################################################################
# START LOGGING CONFIG ###################################################### START LOGGING CONFIG #
####################################################################################################

LOG = logging.getLogger(__name__)


####################################################################################################
# CLASSES ################################################################################ CLASSES #
####################################################################################################


class AccountManager(dds_cli.base.DDSBaseClass):
    """Admin class for adding users, etc."""

    def __init__(
        self, username: str, authenticate: bool = True, method: str = "add", no_prompt: bool = False
    ):
        """Initialize, incl. user authentication."""
        # Initiate DDSBaseClass to authenticate user
        super().__init__(
            username=username, authenticate=authenticate, method=method, no_prompt=no_prompt
        )

        # Only methods "add", "delete" and "revoke" can use the AccountManager class
        if self.method not in ["add", "delete", "revoke"]:
            raise dds_cli.exceptions.AuthenticationError(f"Unauthorized method: '{self.method}'")

    def add_user(self, email, role, project, no_mail=False):
        """Invite new user or associate existing users with projects."""
        # Perform request to API
        json = {"email": email, "role": role, "send_email": not no_mail}

        try:
            response = requests.post(
                dds_cli.DDSEndpoint.USER_ADD,
                headers=self.token,
                params={"project": project},
                json=json,
            )

            # Get response
            response_json = response.json()
            LOG.debug(response_json)
        except requests.exceptions.RequestException as err:
            raise dds_cli.exceptions.ApiRequestError(message=str(err))
        except simplejson.JSONDecodeError as err:
            raise dds_cli.exceptions.ApiResponseError(message=str(err))

        # Format response message
        if not response.ok:
            message = "Could not add user"
            if response.status_code == http.HTTPStatus.INTERNAL_SERVER_ERROR:
                raise dds_cli.exceptions.ApiResponseError(message=f"{message}: {response.reason}")

            raise dds_cli.exceptions.DDSCLIException(
                message=f"{message}: {response_json.get('message', 'Unexpected error!')}"
            )

        dds_cli.utils.console.print(response_json.get("message", "User successfully added."))

    def delete_user(self, email):
        """Delete users from the system"""
        # Perform request to API
        json = {"email": email}

        try:
            response = requests.delete(
                dds_cli.DDSEndpoint.USER_DELETE,
                headers=self.token,
                json=json,
            )

            # Get response
            response_json = response.json()
            message = response_json["message"]

        except requests.exceptions.RequestException as err:
            raise dds_cli.exceptions.ApiRequestError(message=str(err))
        except simplejson.JSONDecodeError as err:
            raise dds_cli.exceptions.ApiResponseError(message=str(err))

        # Format response message
        if not response.ok:
            if response.status_code == http.HTTPStatus.INTERNAL_SERVER_ERROR:
                raise dds_cli.exceptions.ApiResponseError(message)
            else:
                raise dds_cli.exceptions.DDSCLIException(message)
        else:
            LOG.info(message)

    def delete_own_account(self):
        """Delete users from the system"""
        # Perform request to API

        try:
            response = requests.delete(
                dds_cli.DDSEndpoint.USER_DELETE_SELF,
                headers=self.token,
                json=None,
            )

            # Get response
            response_json = response.json()
            message = response_json["message"]

        except requests.exceptions.RequestException as err:
            raise dds_cli.exceptions.ApiRequestError(message=str(err))
        except simplejson.JSONDecodeError as err:
            raise dds_cli.exceptions.ApiResponseError(message=str(err))

        # Format response message
        if not response.ok:
            if response.status_code == http.HTTPStatus.INTERNAL_SERVER_ERROR:
                raise dds_cli.exceptions.ApiResponseError(message)
            else:
                raise dds_cli.exceptions.DDSCLIException(message)
        else:
            LOG.info(message)

    def revoke_project_access(self, project, email):
        """Revoke a user's access to a project"""
        json = {"email": email}
        try:
            response = requests.post(
                dds_cli.DDSEndpoint.REVOKE_PROJECT_ACCESS,
                headers=self.token,
                params={"project": project},
                json=json,
            )

            # Get response
            response_json = response.json()
            LOG.debug(response_json)
        except requests.exceptions.RequestException as err:
            raise dds_cli.exceptions.ApiRequestError(message=str(err))
        except simplejson.JSONDecodeError as err:
            raise dds_cli.exceptions.ApiResponseError(message=str(err))

        # Format response message
        if not response.ok:
            message = "Could not revoke user access"
            if response.status_code == http.HTTPStatus.INTERNAL_SERVER_ERROR:
                raise dds_cli.exceptions.ApiResponseError(message=f"{message}: {response.reason}")

            raise dds_cli.exceptions.DDSCLIException(
                message=f"{message}: {response_json.get('message', 'Unexpected error!')}"
            )

        dds_cli.utils.console.print(
            response_json.get("message", "User access successfully revoked.")
        )

    def get_user_info(self):
        """Get a users info"""
        try:
            response = requests.get(
                dds_cli.DDSEndpoint.DISPLAY_USER_INFO,
                headers=self.token,
            )

            # Get response
            response_json = response.json()
            LOG.debug(response_json)
            info = response_json["info"]
        except requests.exceptions.RequestException as err:
            raise dds_cli.exceptions.ApiRequestError(message=str(err))
        except simplejson.JSONDecodeError as err:
            raise dds_cli.exceptions.ApiResponseError(message=str(err))

        # Format response message
        if not response.ok:
            message = "Could not get user info"
            if response.status_code == http.HTTPStatus.INTERNAL_SERVER_ERROR:
                raise dds_cli.exceptions.ApiResponseError(message=f"{message}: {response.reason}")

            raise dds_cli.exceptions.DDSCLIException(
                message=f"{message}: {response_json.get('message', 'Unexpected error!')}"
            )

        LOG.info(
            f"User Name: {info['username']} \nRole: {info['role']} \
            \nName: {info['name']} \
            \nPrimary Email: {info['email_primary']} \
            \nAssociated Emails: {', '.join(str(x) for x in info['emails_all'])}"
        )

    def user_activation(self, email, action):
        """Deactivate/Reactivate users"""
        json = {"email": email, "action": action}
        try:
            response = requests.post(
                dds_cli.DDSEndpoint.USER_ACTIVATION,
                headers=self.token,
                json=json,
            )

            # Get response
            response_json = response.json()
            LOG.debug(response_json)
        except requests.exceptions.RequestException as err:
            raise dds_cli.exceptions.ApiRequestError(message=str(err))
        except simplejson.JSONDecodeError as err:
            raise dds_cli.exceptions.ApiResponseError(message=str(err))

        # Format response message
        if not response.ok:
            message = f"Could not {action} user"
            if response.status_code == http.HTTPStatus.INTERNAL_SERVER_ERROR:
                raise dds_cli.exceptions.ApiResponseError(message=f"{message}: {response.reason}")

            raise dds_cli.exceptions.DDSCLIException(
                message=f"{message}: {response_json.get('message', 'Unexpected error!')}"
            )

        LOG.info(response_json.get("message", f"User successfully {action}d."))

    def fix_project_access(self, email, project):
        """Fix project access for specific user."""
        json = {"email": email}
        try:
            response = requests.post(
                dds_cli.DDSEndpoint.PROJ_ACCESS,
                headers=self.token,
                params={"project": project},
                json=json,
            )
            response_json = response.json()

        except requests.exceptions.RequestException as err:
            raise dds_cli.exceptions.ApiRequestError(message=str(err))
        except simplejson.JSONDecodeError as err:
            raise dds_cli.exceptions.ApiResponseError(message=str(err))

        if not response.ok:
            message = f"Failed updating user '{email}' project access"
            if response.status_code == http.HTTPStatus.INTERNAL_SERVER_ERROR:
                raise dds_cli.exceptions.ApiResponseError(message=f"{message}: {response.reason}")

            raise dds_cli.exceptions.DDSCLIException(
                message=f"{message}: {response_json.get('message', 'Unexpected error!')}"
            )

        LOG.info(
            response_json.get(
                "message",
                (
                    f"Project access fixed for user '{email}'. "
                    "They should now have access to all project data."
                ),
            )
        )



"""Local file handler module"""

###############################################################################
# IMPORTS ########################################################### IMPORTS #
###############################################################################

# Standard library
import hashlib
import http
import logging
import os
import pathlib
import uuid
import random

# Installed
import requests
import simplejson

# Own modules
from dds_cli import DDSEndpoint
from dds_cli import file_compressor as fc
from dds_cli import file_handler as fh
from dds_cli import FileSegment
from dds_cli import exceptions
import dds_cli.utils

###############################################################################
# START LOGGING CONFIG ################################# START LOGGING CONFIG #
###############################################################################

LOG = logging.getLogger(__name__)

###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


class LocalFileHandler(fh.FileHandler):
    """Collects the files specified by the user."""

    # Magic methods ################ Magic methods #
    def __init__(self, user_input, temporary_destination, project):

        LOG.debug("Collecting file info...")

        # Initiate FileHandler from inheritance
        super().__init__(
            user_input=user_input, local_destination=temporary_destination, project=project
        )

        # Remove duplicates and save all files for later use
        all_files = set(self.data_list)

        # Remove non existent files
        self.data_list = {x for x in self.data_list if pathlib.Path(x).exists()}

        non_existent_files = all_files.difference(self.data_list)
        if len(non_existent_files) > 0:
            # Issue warning that some of the files don't exist
            LOG.warning(
                "The following files from '{}' does not exist: '{}'".format(
                    user_input[1], "', '".join(non_existent_files)
                )
            )

        # Get absolute paths for all data
        self.data_list = [
            pathlib.Path(os.path.abspath(os.path.expanduser(path))) for path in self.data_list
        ]

        # No data -- cannot proceed
        if not self.data_list:
            dds_cli.utils.console.print("\n:warning-emoji: No data specified. :warning-emoji:\n")
            os._exit(1)

        self.data, _ = self.__collect_file_info_local(all_paths=self.data_list)
        self.data_list = None

        LOG.debug("File info computed/collected")

    # Static methods ############## Static methods #
    @staticmethod
    def generate_bucket_filepath(filename="", folder=pathlib.Path("")):
        """Generates filename and new path which the file will be
        called in the bucket."""

        # Generate new file name
        new_name = f"{'%020x' % random.randrange(16**20)}_{uuid.uuid5(uuid.NAMESPACE_X500, str(folder))}{uuid.uuid5(uuid.NAMESPACE_X500, filename)}"
        return new_name

    @staticmethod
    def read_file(file, chunk_size: int = FileSegment.SEGMENT_SIZE_RAW):
        """Read file in chunk_size sized chunks."""

        try:
            with file.open(mode="rb") as infile:
                for chunk in iter(lambda: infile.read(chunk_size), b""):
                    yield chunk
        except OSError as err:
            LOG.warning(str(err))

    # Private methods ############ Private methods #
    def __collect_file_info_local(self, all_paths, folder=pathlib.Path(""), task_name=""):
        """Get info on each file in each path specified."""

        file_info = dict()
        progress_tasks = dict()
        for path in all_paths:
            task_name = path.name if folder == pathlib.Path("") else task_name
            # Get info for all files
            # and feed back to same function for all folders
            if path.is_file():
                with fc.Compressor() as compressor:
                    is_compressed, error = compressor.is_compressed(file=path)

                    if error != "":
                        LOG.exception(error)
                        os._exit(1)

                path_processed = self.create_encrypted_name(
                    raw_file=path,
                    subpath=folder,
                    no_compression=is_compressed,
                )

                file_info[str(folder / path.name)] = {
                    "path_raw": path,
                    "subpath": folder,
                    "size_raw": path.stat().st_size,
                    "compressed": is_compressed,
                    "path_processed": path_processed,
                    "size_processed": 0,
                    "path_remote": self.generate_bucket_filepath(
                        filename=path_processed.name, folder=folder
                    ),
                    "overwrite": False,
                    "checksum": "",
                }

            elif path.is_dir():
                content_info, _ = self.__collect_file_info_local(
                    all_paths=path.glob("*"),
                    folder=folder / pathlib.Path(path.name),
                )
                file_info.update({**content_info})
            else:
                if path.is_symlink():
                    try:
                        resolved = path.resolve()
                    except RuntimeError:
                        LOG.warning(
                            f"IGNORED: Link: {path} seems to contain infinite loop, will be ignored."
                        )
                    else:
                        LOG.warning(
                            f"IGNORED: Link: {path} -> {resolved} seems to be broken, will be ignored."
                        )
                else:
                    LOG.warning(
                        f"IGNORED: Path of unsupported/unknown type: {path}, will be ignored."
                    )

        return file_info, progress_tasks

    # Public methods ############## Public methods #
    def create_upload_status_dict(self, existing_files, overwrite=False):
        """Create dict for tracking file delivery status"""

        LOG.debug("Creating the status dictionary.")

        status_dict = {}
        for x in list(self.data):
            in_db = bool(x in existing_files)
            if in_db and not overwrite:
                self.failed[x] = {
                    **self.data.pop(x),
                    **{"message": "File already uploaded"},
                }
            else:
                if in_db:
                    if overwrite:
                        self.data[x].update(
                            {
                                "overwrite": True,
                                "path_remote": existing_files[x],
                            }
                        )

                # filestream_funcname = (
                #     "read_file" if self.data[x]["compressed"] else "compress_file"
                # )
                status_dict[x] = {
                    "cancel": False,
                    "started": False,
                    "message": "",
                    "failed_op": None,
                    # filestream_funcname: {"started": False, "done": False},
                    "put": {"started": False, "done": False},
                    "add_file_db": {"started": False, "done": False},
                    # "task": None,
                }

        LOG.debug("Initial statuses created.")

        return status_dict

    def check_previous_upload(self, token):
        """Do API call and check for the files in the DB."""

        LOG.debug("Checking if files have been previously uploaded.")

        # Get files from db
        files = list(x for x in self.data)
        try:
            response = requests.get(
                DDSEndpoint.FILE_MATCH,
                params={"project": self.project},
                headers=token,
                json=files,
                timeout=DDSEndpoint.TIMEOUT,
            )
        except requests.exceptions.RequestException as err:
            LOG.warning(err)
            raise SystemExit from err

        if not response.ok:
            message = "Failed getting information about previously uploaded files"
            if response.status_code == http.HTTPStatus.INTERNAL_SERVER_ERROR:
                raise exceptions.ApiResponseError(message=f"{message}: {response.reason}")

            raise exceptions.DDSCLIException(message=f"{message}: {response.json().get('message')}")

        try:
            files_in_db = response.json()
        except simplejson.JSONDecodeError as err:
            LOG.warning(err)
            raise SystemExit from err

        # API failure
        if "files" not in files_in_db:
            dds_cli.utils.console.print(
                "\n:warning-emoji: Files not returned from API. :warning-emoji:\n"
            )
            os._exit(1)

        LOG.debug("Previous upload check finished.")

        return dict() if files_in_db["files"] is None else files_in_db["files"]

    def create_encrypted_name(
        self, raw_file: pathlib.Path, subpath: str = "", no_compression: bool = True
    ):
        """Create new file name to save encrypted file."""

        # New local file name
        old_suffix = "".join(raw_file.suffixes)
        new_suffix = old_suffix if no_compression else f"{old_suffix}.zst"
        new_file_name = (
            self.local_destination / subpath / raw_file.with_suffix(new_suffix + ".ccp").name
        )

        return new_file_name

    def stream_from_file(self, file):
        """Read raw or compress file depending on if compressed already or not."""

        file_info = self.data[file]

        LOG.debug("Streaming file...")
        # Generate checksum
        checksum = hashlib.sha256()
        if file_info["compressed"]:
            for chunk in self.read_file(file=file_info["path_raw"]):
                checksum.update(chunk)
                yield chunk
        else:
            LOG.debug("File not compressed -- compressing")
            # Generate checksum first
            # total_read = 0
            for chunk in self.read_file(file=file_info["path_raw"]):
                checksum.update(chunk)
                # total_read += len(chunk)
                # LOG.debug(total_read)

            # Then stream file chunks
            # LOG.debug(
            #     "Test: %s",
            # )
            for chunk in fc.Compressor.compress_file(file=file_info["path_raw"]):
                # LOG.debug("Chunk type: %s", type(chunk))
                # checksum.update(chunk)
                # break
                yield chunk

        # LOG.debug("Streaming file finished.")
        # Add checksum to file info
        self.data[file]["checksum"] = checksum.hexdigest()



"""Remote file handler module."""

###############################################################################
# IMPORTS ########################################################### IMPORTS #
###############################################################################

# Standard library
import logging
import os
import pathlib

# Installed
import requests
import simplejson

# Own modules
from dds_cli import DDSEndpoint
from dds_cli import file_handler as fh
import dds_cli.utils

###############################################################################
# START LOGGING CONFIG ################################# START LOGGING CONFIG #
###############################################################################

LOG = logging.getLogger(__name__)

###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


class RemoteFileHandler(fh.FileHandler):
    """Collects the files specified by the user."""

    # Magic methods ################ Magic methods #
    def __init__(self, get_all, user_input, token, project, destination=pathlib.Path("")):
        """Initialize FileHandler for collecting remote files."""
        # Initiate FileHandler from inheritance
        super().__init__(user_input=user_input, local_destination=destination, project=project)

        self.get_all = get_all

        self.data_list = list(set(self.data_list))

        if not self.data_list and not get_all:
            raise dds_cli.exceptions.NoDataError(
                "\n:warning-emoji: No data specified. :warning-emoji:\n"
            )

        self.data = self.__collect_file_info_remote(all_paths=self.data_list, token=token)
        self.data_list = None

    # Static methods ############ Static methods #
    @staticmethod
    def write_file(chunks, outfile: pathlib.Path, **_):
        """Write file chunks to file."""
        saved, message = (False, "")

        LOG.debug("Saving file...")
        try:
            with outfile.open(mode="wb+") as new_file:
                for chunk in chunks:
                    new_file.write(chunk)
        except OSError as err:
            message = str(err)
            LOG.exception(message)
        else:
            saved = True
            LOG.debug("File saved.")

        return saved, message

    # Private methods ############ Private methods #
    def __collect_file_info_remote(self, all_paths, token):
        """Get information on files in db."""
        # Get file info from db via API
        try:
            response = requests.get(
                DDSEndpoint.FILE_INFO_ALL if self.get_all else DDSEndpoint.FILE_INFO,
                params={"project": self.project},
                headers=token,
                json=all_paths,
            )

            # Get file info from response
            file_info = response.json()
        except requests.exceptions.RequestException as err:
            raise dds_cli.exceptions.ApiRequestError(message=str(err))
        except simplejson.JSONDecodeError as err:
            raise dds_cli.exceptions.ApiResponseError(message=str(err))

        # Server error or error in response
        if not response.ok:
            raise dds_cli.exceptions.ApiResponseError(response.text)

        # Folder info required if specific files requested
        if all_paths and not all(x in file_info for x in ["files", "folder_contents", "not_found"]):
            raise dds_cli.exceptions.DDSCLIException(
                "Error in response. Not enough info returned despite ok request."
            )

        folder_contents = file_info.get("folder_contents", {})
        files = file_info.get("files")

        LOG.debug(f"Attempted: \n{all_paths}")
        LOG.debug(f"Files: \n{files}")
        LOG.debug(f"Folder contents: \n{folder_contents}")

        # Cancel download of those files or folders not found in the db
        self.failed = {
            x: {"error": "Not found in DB."}
            for x in all_paths
            if x not in files and x not in folder_contents
        }

        LOG.debug(f"Not found: {self.failed}")

        # Save info on files in dict and return
        data = {
            self.local_destination
            / pathlib.Path(x): {
                **y,
                "name_in_db": x,
                "path_downloaded": self.local_destination
                / pathlib.Path(y["subpath"])
                / pathlib.Path(y["name_in_bucket"]),
            }
            for x, y in files.items()
        }
        LOG.debug(f"Data (files):\n {data}")

        # Save info on files in a specific folder and return
        for x, y in folder_contents.items():
            LOG.debug(f"{x}")
            data.update(
                {
                    self.local_destination
                    / pathlib.Path(j): {
                        **k,
                        "name_in_db": j,
                        "path_downloaded": self.local_destination
                        / pathlib.Path(k["subpath"])
                        / pathlib.Path(k["name_in_bucket"]),
                    }
                    for j, k in y.items()
                }
            )

        LOG.debug(f"Data (files and folders):\n {data}")
        return data

    # Public methods ############ Public methods #
    def create_download_status_dict(self):
        """Create dict for tracking file download status."""
        return {
            x: {
                "cancel": False,
                "started": False,
                "message": "",
                "failed_op": None,
                "get": {"started": False, "done": False},
                "update_db": {"started": False, "done": False},
            }
            for x in self.data
        }



# Standard library
import unittest.mock

# Installed
import click.testing
import pytest

# Own modules
from dds_cli.user import User
from dds_cli.__main__ import dds_main


@pytest.fixture
def retrieve_token():
    """Fixture to mock authentication by having a None token for every user."""
    with unittest.mock.patch.object(User, "_User__retrieve_token") as mock_A:
        mock_A.return_value = None
        yield mock_A


@pytest.fixture
def runner(retrieve_token):
    """
    Fixture that returns the click cli runner.

    The runner is invoked when the function returned by this fixture is called.
    """
    runner_ = click.testing.CliRunner(mix_stderr=False)

    def _run(cmd, input=None):
        return runner_.invoke(
            dds_main,
            cmd,
            catch_exceptions=True,
            input=input,
        )

    yield _run



"""CLI for the Data Delivery System."""

####################################################################################################
# IMPORTS ################################################################################ IMPORTS #
####################################################################################################

# Standard library
import concurrent.futures
import itertools
import logging
import os
import sys

# Installed
import rich_click as click
import click_pathlib
import rich
import rich.logging
import rich.progress
import rich.prompt
import questionary

# Own modules
import dds_cli
import dds_cli.account_manager
import dds_cli.data_getter
import dds_cli.data_lister
import dds_cli.data_putter
import dds_cli.data_remover
import dds_cli.directory
import dds_cli.project_creator
import dds_cli.auth
import dds_cli.project_status
import dds_cli.user
import dds_cli.utils
from dds_cli.options import (
    email_arg,
    email_option,
    folder_option,
    num_threads_option,
    project_option,
    sort_projects_option,
    source_option,
    source_path_file_option,
    username_option,
    break_on_fail_flag,
    json_flag,
    nomail_flag,
    silent_flag,
    size_flag,
    tree_flag,
    usage_flag,
    users_flag,
)

####################################################################################################
# START LOGGING CONFIG ###################################################### START LOGGING CONFIG #
####################################################################################################

LOG = logging.getLogger()

# Configuration for rich-click output
click.rich_click.MAX_WIDTH = 100


## # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                                                                                                  #
#                          MMMM   MMMM      AAAA      II   NNNN    NN                              #
#                          MM MM MM MM     AA  AA     II   NN NN   NN                              #
#                          MM  MMM  MM    AA    AA    II   NN  NN  NN                              #
#                          MM   M   MM   AAAAAAAAAA   II   NN   NN NN                              #
#                          MM       MM   AA      AA   II   NN    NNNN                              #
#                                                                                                  #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # ##

# Get token metadata
username = dds_cli.user.User.get_user_name_if_logged_in()

# Print header to STDERR
dds_cli.utils.stderr_console.print(
    "[green]     ︵",
    "\n[green] ︵ (  )   ︵",
    "\n[green](  ) ) (  (  )[/]   [bold]SciLifeLab Data Delivery System",
    "\n[green] ︶  (  ) ) ([/]    [blue][link={0}]{0}[/link]".format(dds_cli.__url__),
    f"\n[green]      ︶ (  )[/]    [dim]Version {dds_cli.__version__}",
    "\n[green]          ︶",
    f"\n[green]Current user:[/] [red]{username}" if username else "",
    highlight=False,
)

# -- dds -- #
@click.group()
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="Print verbose output to the console."
)
@click.option("-l", "--log-file", help="Save a log to a file.", metavar="<filename>")
@click.option(
    "--no-prompt", is_flag=True, default=False, help="Run without any interactive features."
)
@click.version_option(version=dds_cli.__version__, prog_name=dds_cli.__title__)
@click.pass_context
def dds_main(click_ctx, verbose, log_file, no_prompt):
    """SciLifeLab Data Delivery System (DDS) command line interface.

    Access token is saved in a .dds_cli_token file in the home directory.
    """
    if "--help" not in sys.argv:
        # Set the base logger to output DEBUG
        LOG.setLevel(logging.DEBUG)

        # Set up logs to the console
        LOG.addHandler(
            rich.logging.RichHandler(
                level=logging.DEBUG if verbose else logging.INFO,
                console=dds_cli.utils.stderr_console,
                show_time=False,
                markup=True,
                show_path=verbose,
            )
        )

        # Set up logs to a file if we asked for one
        if log_file:
            log_fh = logging.FileHandler(log_file, encoding="utf-8")
            log_fh.setLevel(logging.DEBUG)
            log_fh.setFormatter(
                logging.Formatter("[%(asctime)s] %(name)-20s [%(levelname)-7s]  %(message)s")
            )
            LOG.addHandler(log_fh)

        # Create context object
        click_ctx.obj = {"NO_PROMPT": no_prompt}


# ************************************************************************************************ #
# MAIN DDS COMMANDS ************************************************************ MAIN DDS COMMANDS #
# ************************************************************************************************ #

# -- dds ls -- #
@dds_main.command(name="ls")
# Options
@username_option()
@project_option(required=False)
@sort_projects_option()
@folder_option(help_message="List contents of this project folder.")
# Flags
@json_flag(help_message="Output in JSON format.")
@size_flag(help_message="Show size of project contents.")
@tree_flag(help_message="Display the entire project(s) directory tree.")
@usage_flag(help_message="Show the usage for available projects, in GBHours and cost.")
@users_flag(help_message="Display users associated with a project(Requires a project id).")
@click.option("--projects", "-lp", is_flag=True, help="List all project connected to your account.")
@click.pass_obj
def list_projects_and_contents(
    click_ctx, project, folder, username, sort, json, size, tree, usage, users, projects
):
    """List the projects you have access to or the project contents.

    To list all projects, run `dds ls` without any arguments.

    Specify a Project ID to list the files within a project.
    You can also follow this with a subfolder path to show files within that folder.
    """
    try:
        # List all projects if project is None and all files if project spec
        if project is None:
            with dds_cli.data_lister.DataLister(
                project=project,
                show_usage=usage,
                username=username,
                no_prompt=click_ctx.get("NO_PROMPT", False),
                json=json,
            ) as lister:
                projects = lister.list_projects(sort_by=sort)
                if json:
                    dds_cli.utils.console.print_json(data=projects)
                else:
                    # If an interactive terminal, ask user if they want to view files for a project
                    if sys.stdout.isatty() and not lister.no_prompt:
                        project_ids = [p["Project ID"] for p in projects]
                        LOG.info(
                            "Would you like to view files in a specific project? "
                            "Leave blank to exit."
                        )
                        # Keep asking until we get a valid response
                        while project not in project_ids:
                            try:
                                project = questionary.autocomplete(
                                    "Project ID:",
                                    choices=project_ids,
                                    validate=lambda x: x in project_ids or x == "",
                                    style=dds_cli.dds_questionary_styles,
                                ).unsafe_ask()
                                assert project and project != ""

                            # If didn't enter anything, convert to None and exit
                            except (KeyboardInterrupt, AssertionError):
                                break

        # List all files in a project if we know a project ID
        if project:
            with dds_cli.data_lister.DataLister(
                project=project,
                username=username,
                tree=tree,
                no_prompt=click_ctx.get("NO_PROMPT", False),
                json=json,
            ) as lister:
                if json:
                    json_output = {"project_name": project}
                    if users:
                        user_list = lister.list_users()
                        json_output["users"] = user_list

                    if tree:
                        folders = lister.list_recursive(show_size=size)
                        json_output["project_files_and_directories"] = folders
                    else:
                        LOG.warning(
                            "JSON output for file listing only possible for the complete file tree."
                            " Please use the '--tree' option to view complete contens in JSON or "
                            "remove the '--json' option to list files interactively"
                        )
                    dds_cli.utils.console.print_json(data=json_output)
                else:
                    if users:
                        user_list = lister.list_users()

                    if tree:
                        folders = lister.list_recursive(show_size=size)
                    else:
                        folders = lister.list_files(folder=folder, show_size=size)

                        # If an interactive terminal, ask user if they want to view files for a proj
                        if sys.stdout.isatty() and (not lister.no_prompt) and len(folders) > 0:
                            LOG.info(
                                "Would you like to view files within a directory? "
                                "Leave blank to exit."
                            )
                            last_folder = None
                            while folder is None or folder != last_folder:
                                last_folder = folder

                                try:
                                    folder = questionary.autocomplete(
                                        "Folder:",
                                        choices=folders,
                                        validate=lambda x: x in folders or x == "",
                                        style=dds_cli.dds_questionary_styles,
                                    ).unsafe_ask()
                                    assert folder != ""
                                    assert folder is not None
                                # If didn't enter anything, convert to None and exit
                                except (KeyboardInterrupt, AssertionError):
                                    break

                                # Prepend existing file path
                                if last_folder is not None and folder is not None:
                                    folder = os.path.join(last_folder, folder)

                                # List files
                                folders = lister.list_files(folder=folder, show_size=size)

                                if len(folders) == 0:
                                    break

    except (dds_cli.exceptions.NoDataError) as err:
        LOG.warning(err)
        sys.exit(0)
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


####################################################################################################
####################################################################################################
## AUTH #################################################################################### AUTH ##
####################################################################################################
####################################################################################################


@dds_main.group(name="auth", no_args_is_help=True)
@click.pass_obj
def auth_group_command(_):
    """Group command for creating and managing authenticated sessions.

    Authenticate yourself once and run multiple commands within a certain amount of time
    (currently 48 hours) without specifying your user credentials.
    If you do not authenticate yourself and start a new session, you will need to provide your
    DDS username when running the other commands. If you do not provide the username as an option,
    you will be prompted to fill it in.
    """


# ************************************************************************************************ #
# AUTH COMMANDS ******************************************************************** AUTH COMMANDS #
# ************************************************************************************************ #


# -- dds auth login -- #
@auth_group_command.command(name="login")
@username_option()
@click.pass_obj
def login(click_ctx, username):
    """Start or renew an authenticated session.

    Creates or renews the authentication token stored in the '.dds_cli_token' file.

    Run this command before running the cli in a non-interactive fashion as this enables the longest
    possible session time before a password needs to be entered again.
    """
    no_prompt = click_ctx.get("NO_PROMPT", False)
    if no_prompt:
        LOG.warning("The --no-prompt flag is ignored for `dds auth login`")
    try:
        with dds_cli.auth.Auth(username=username):
            # Authentication token renewed in the init method.
            LOG.info("[green] :white_check_mark: Authentication token renewed![/green]")
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds auth logout -- #
@auth_group_command.command(name="logout")
def logout():
    """End authenticated session.

    Removes the saved authentication token by deleting the '.dds_cli_token' file.
    """
    try:
        with dds_cli.auth.Auth(username=None, authenticate=False) as authenticator:
            authenticator.logout()
            LOG.info("[green]Successfully logged out![/green]")
    except dds_cli.exceptions.DDSCLIException as err:
        LOG.error(err)
        sys.exit(1)


# -- dds auth info -- #
@auth_group_command.command(name="info")
def info():
    """Display information about token.

    Information displayed:
    - If the token is about to expire
    - Token age
    - Time of token expiration
    """
    try:
        with dds_cli.auth.Auth(username=None, authenticate=False) as authenticator:
            authenticator.check()
    except dds_cli.exceptions.DDSCLIException as err:
        LOG.error(err)
        sys.exit(1)


####################################################################################################
####################################################################################################
## USER #################################################################################### USER ##
####################################################################################################
####################################################################################################


@dds_main.group(name="user", no_args_is_help=True)
@click.pass_obj
def user_group_command(_):
    """Group command for managing user accounts, including your own."""


# ************************************************************************************************ #
# USER COMMANDS ******************************************************************** USER COMMANDS #
# ************************************************************************************************ #

# -- dds user add -- #
@user_group_command.command(name="add", no_args_is_help=True)
# Positional args
@email_arg(required=True)
# Options
@username_option()
@project_option(
    required=False, help_message="Existing Project you want the user to be associated to."
)
@click.option(
    "--role",
    "-r",
    "role",
    required=True,
    type=click.Choice(
        choices=["Super Admin", "Unit Admin", "Unit Personnel", "Project Owner", "Researcher"],
        case_sensitive=False,
    ),
    help="Type of account.",
)
@nomail_flag(help_message="Do not send e-mail notifications regarding project updates.")
@click.pass_obj
def add_user(click_ctx, email, username, role, project, no_mail):
    """
    Add a user to the DDS system or hosted projects.

    Specify a users email and role to associate it with projects.

    If the user doesn't exist in the system yet, an invitation email
    will be sent automatically to that person.
    """
    try:
        with dds_cli.account_manager.AccountManager(
            username=username, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as inviter:
            inviter.add_user(email=email, role=role, project=project, no_mail=no_mail)
    except (
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.ApiResponseError,
        dds_cli.exceptions.ApiRequestError,
        dds_cli.exceptions.DDSCLIException,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds user delete -- #
@user_group_command.command(name="delete", no_args_is_help=True)
# Positional args
@email_arg(required=False)
# Options
@username_option()
# Flags
@click.option(
    "--self",
    "self",
    required=False,
    is_flag=True,
    default=False,
    help="Request deletion of own account.",
)
@click.pass_obj
def delete_user(click_ctx, email, username, self):
    """
    Delete user accounts from the Data Delivery System.

    Use this command with caution. Deletion of accounts cannot be undone.

    To request the removal of your own account, use the `--self` flag without any arguments.
    An e-mail will be sent to you asking to confirm the deletion.

    If you have sufficient admin privileges, you may also delete the accounts of other users.
    Specify the e-mail address as argument to the main command to initiate the removal process.
    """
    if click_ctx.get("NO_PROMPT", False):
        pass
    else:
        if not self:
            proceed_deletion = rich.prompt.Confirm.ask(
                f"Delete Data Delivery System user account associated with {email}"
            )
        else:
            proceed_deletion = rich.prompt.Confirm.ask(
                "Are you sure? Deleted accounts can't be restored!"
            )

    if proceed_deletion:
        try:
            with dds_cli.account_manager.AccountManager(
                username=username,
                method="delete",
                no_prompt=click_ctx.get("NO_PROMPT", False),
            ) as manager:
                if self and not email:
                    manager.delete_own_account()
                elif email and not self:
                    manager.delete_user(email=email)
                else:
                    LOG.error(
                        "You must either specify the '--self' flag "
                        "or the e-mail address of the user to be deleted"
                    )
                    sys.exit(1)

        except (
            dds_cli.exceptions.AuthenticationError,
            dds_cli.exceptions.ApiResponseError,
            dds_cli.exceptions.ApiRequestError,
            dds_cli.exceptions.DDSCLIException,
        ) as err:
            LOG.error(err)
            sys.exit(1)


# -- dds user info -- #
@user_group_command.command(name="info")
# Options
@username_option()
# Flags
@click.pass_obj
def get_info_user(click_ctx, username):
    """Display information connected to your own DDS account."""
    try:
        with dds_cli.account_manager.AccountManager(
            username=username, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as get_info:
            get_info.get_user_info()
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds user activate -- #
@user_group_command.command(name="activate", no_args_is_help=True)
# Positional args
@email_arg(required=True)
# Options
@username_option()
# Flags
@click.pass_obj
def activate_user(click_ctx, email, username):
    """Activate/Reactivate user accounts.

    If you have sufficient admin privileges, you may activate the accounts of other users.
    Specify the e-mail address as argument to the main command to initiate the activation process.
    """
    if click_ctx.get("NO_PROMPT", False):
        pass
    else:
        proceed_activation = rich.prompt.Confirm.ask(
            f"Activate Data Delivery System user account associated with {email}?"
        )

    if proceed_activation:
        try:
            with dds_cli.account_manager.AccountManager(
                username=username,
                no_prompt=click_ctx.get("NO_PROMPT", False),
            ) as manager:
                manager.user_activation(email=email, action="reactivate")

        except (
            dds_cli.exceptions.AuthenticationError,
            dds_cli.exceptions.ApiResponseError,
            dds_cli.exceptions.ApiRequestError,
            dds_cli.exceptions.DDSCLIException,
        ) as err:
            LOG.error(err)
            sys.exit(1)


# -- dds user deactivate -- #
@user_group_command.command(name="deactivate", no_args_is_help=True)
# Positional args
@email_arg(required=True)
# Options
@username_option()
# Flags
@click.pass_obj
def deactivate_user(click_ctx, email, username):
    """Deactivate user accounts in the Data Delivery System.

    If you have sufficient admin privileges, you may deactivate the accounts of other users.
    Specify the e-mail address as argument to the main command to initiate the deactivation process.
    """
    if click_ctx.get("NO_PROMPT", False):
        pass
    else:
        proceed_deactivation = rich.prompt.Confirm.ask(
            f"Deactivate Data Delivery System user account associated with {email}?"
        )

    if proceed_deactivation:
        try:
            with dds_cli.account_manager.AccountManager(
                username=username,
                no_prompt=click_ctx.get("NO_PROMPT", False),
            ) as manager:
                manager.user_activation(email=email, action="deactivate")

        except (
            dds_cli.exceptions.AuthenticationError,
            dds_cli.exceptions.ApiResponseError,
            dds_cli.exceptions.ApiRequestError,
            dds_cli.exceptions.DDSCLIException,
        ) as err:
            LOG.error(err)
            sys.exit(1)


####################################################################################################
####################################################################################################
## PROJECT ############################################################################## PROJECT ##
####################################################################################################
####################################################################################################


@dds_main.group(name="project", no_args_is_help=True)
@click.pass_obj
def project_group_command(_):
    """Group command for creating and managing projects within the DDS."""


# ************************************************************************************************ #
# PROJECT COMMANDS ************************************************************** PROJECT COMMANDS #
# ************************************************************************************************ #


# -- dds project ls -- #
@project_group_command.command(name="ls")
# Options
@username_option()
@sort_projects_option()
# Flags
@usage_flag(help_message="Show the usage for available projects, in GBHours and cost.")
@json_flag(help_message="Output project list as json.")  # users, json, tree
@click.pass_context
def list_projects(ctx, username, json, sort, usage):
    """List all projects you have access to in the DDS.

    Calls the `dds ls` function.
    """
    ctx.invoke(list_projects_and_contents, username=username, json=json, sort=sort, usage=usage)


# -- dds project create -- #
@project_group_command.command(no_args_is_help=True)
# Options
@username_option()
@click.option(
    "--title",
    "-t",
    required=True,
    type=str,
    help="The title of the project.",
)
@click.option(
    "--description",
    "-d",
    required=True,
    type=str,
    help="A description of the project.",
)
@click.option(
    "--principal-investigator",
    "-pi",
    required=True,
    type=str,
    help="The name of the Principal Investigator.",
)
@click.option(
    "--researcher",
    required=False,
    multiple=True,
    help="Email of a user to be added to the project as Researcher."
    + dds_cli.utils.multiple_help_text(item="researcher"),
)
# Flags
@click.option(
    "--owner",
    "owner",
    required=False,
    multiple=True,
    help="Email of user to be added to the project as Project Owner."
    + dds_cli.utils.multiple_help_text(item="project owner"),
)
@click.option(
    "--non-sensitive",
    required=False,
    is_flag=True,
    help="Indicate whether the project contains only non-sensitive data",
)
@click.pass_obj
def create(
    click_ctx,
    username,
    title,
    description,
    principal_investigator,
    non_sensitive,
    owner,
    researcher,
):
    """Create a project."""
    try:
        with dds_cli.project_creator.ProjectCreator(
            username=username, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as creator:
            emails_roles = []
            if owner or researcher:
                email_overlap = set(owner) & set(researcher)
                if email_overlap:
                    LOG.info(
                        f"The email(s) {email_overlap} specified as both owner and researcher! "
                        "Please specify a unique role for each email."
                    )
                    sys.exit(1)
                if owner:
                    emails_roles.extend([{"email": x, "role": "Project Owner"} for x in owner])
                if researcher:
                    emails_roles.extend([{"email": x, "role": "Researcher"} for x in researcher])

            created, project_id, user_addition_messages, err = creator.create_project(
                title=title,
                description=description,
                principal_investigator=principal_investigator,
                non_sensitive=non_sensitive,
                users_to_add=emails_roles,
            )
            if created:
                dds_cli.utils.console.print(
                    f"Project created with id: {project_id}",
                )
                if user_addition_messages:
                    for msg in user_addition_messages:
                        dds_cli.utils.console.print(msg)
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# ************************************************************************************************ #
# PROJECT SUB GROUPS ********************************************************** PROJECT SUB GROUPS #
# ************************************************************************************************ #


# STATUS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ STATUS #
@project_group_command.group(name="status", no_args_is_help=True)
@click.pass_obj
def project_status(_):
    """Manage project statuses."""


# -- dds project status display -- #
@project_status.command(name="display", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
# Flags
@click.option(
    "--show_history",
    required=False,
    is_flag=True,
    help="Show history of project statuses in addition to current status.",
)
@click.pass_obj
def display_project_status(click_ctx, username, project, show_history):
    """Display and manage project statuses."""
    try:
        with dds_cli.project_status.ProjectStatusManager(
            username=username, project=project, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as updater:
            updater.get_status(show_history)
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds project status release -- #
@project_status.command(name="release", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
@click.option(
    "--deadline",
    required=False,
    type=int,
    help="Deadline in days when releasing a project.",
)
@nomail_flag(help_message="Do not send e-mail notifications regarding project updates.")
@click.pass_obj
def release_project(click_ctx, username, project, deadline, no_mail):
    """Make project data available for user download."""
    try:
        with dds_cli.project_status.ProjectStatusManager(
            username=username, project=project, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as updater:
            updater.update_status(new_status="Available", deadline=deadline, no_mail=no_mail)
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds project status retract -- #
@project_status.command(name="retract", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
@click.pass_obj
def retract_project(click_ctx, username, project):
    """Set the status as `In Progress`.

    This allows Unit Personnel / Admins to upload additional data to the project.
    """
    try:
        with dds_cli.project_status.ProjectStatusManager(
            username=username, project=project, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as updater:
            updater.update_status(new_status="In Progress")
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds project status archive -- #
@project_status.command(name="archive", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
@click.pass_obj
def archive_project(click_ctx, username, project):
    """Manually archive a released project.

    This deletes all project data.
    """
    try:
        with dds_cli.project_status.ProjectStatusManager(
            username=username, project=project, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as updater:
            updater.update_status(new_status="Archived")
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds project status delete -- #
@project_status.command(name="delete", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
@click.pass_obj
def delete_project(click_ctx, username, project):
    """Delete an unreleased project.

    This deletes all project data.
    """
    try:
        with dds_cli.project_status.ProjectStatusManager(
            username=username, project=project, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as updater:
            updater.update_status(new_status="Deleted")
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds project status abort -- #
@project_status.command(name="abort", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
@click.pass_obj
def abort_project(click_ctx, username, project):
    """Abort a released project.

    This deletes all project data.
    """
    try:
        with dds_cli.project_status.ProjectStatusManager(
            username=username, project=project, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as updater:
            updater.update_status(new_status="Archived", is_aborted=True)
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# ACCESS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ ACCESS #
@project_group_command.group(name="access")
@click.pass_obj
def project_access(_):
    """Manage specific users access to a project."""


# -- dds project access grant -- #
@project_access.command(name="grant", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
@email_option(help_message="Email of the user you would like to grant access to the project.")
# Flags
@click.option(
    "--owner",
    "owner",
    required=False,
    is_flag=True,
    help="Grant access as project owner. If not specified, "
    "the user gets Researcher permissions within the project.",
)
@nomail_flag(help_message="Do not send e-mail notifications regarding project updates.")
@click.pass_obj
def grant_project_access(click_ctx, username, project, email, owner, no_mail):
    """Grant user access to a project."""
    try:
        with dds_cli.account_manager.AccountManager(
            username=username, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as granter:
            role = "Researcher"
            if owner:
                role = "Project Owner"
            granter.add_user(email=email, role=role, project=project, no_mail=no_mail)
    except (
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.ApiResponseError,
        dds_cli.exceptions.ApiRequestError,
        dds_cli.exceptions.DDSCLIException,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds project access revoke -- #
@project_access.command(name="revoke", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
@email_option(help_message="Email of the user for whom project access is to be revoked.")
@click.pass_obj
def revoke_project_access(click_ctx, username, project, email):
    """Revoke user access to a project."""
    try:
        with dds_cli.account_manager.AccountManager(
            username=username, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as revoker:
            revoker.revoke_project_access(project, email)
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds project access fix -- #
@project_access.command(name="fix", no_args_is_help=True)
# Positional arguments
@email_arg(required=True)
# Options
@username_option()
@project_option(required=False)
@click.pass_obj
def fix_project_access(click_ctx, email, username, project):
    """Re-grant project access to user that has lost access due to password reset."""
    try:
        with dds_cli.account_manager.AccountManager(
            username=username, no_prompt=click_ctx.get("NO_PROMPT", False)
        ) as fixer:
            fixer.fix_project_access(email=email, project=project)
    except (
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


####################################################################################################
####################################################################################################
## DATA #################################################################################### DATA ##
####################################################################################################
####################################################################################################


@dds_main.group(name="data", no_args_is_help=True)
@click.pass_obj
def data_group_command(_):
    """Group command for uploading, downloading and managing project data."""


# ************************************************************************************************ #
# DATA COMMANDS ******************************************************************** DATA COMMANDS #
# ************************************************************************************************ #


# -- dds data put -- #
@data_group_command.command(name="put", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True, help_message="Project ID to which you're uploading data.")
@source_option(
    help_message="Path to file or directory (local).", option_type=click.Path(exists=True)
)
@source_path_file_option()
@num_threads_option()
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    show_default=True,
    help="Overwrite files if already uploaded.",
)
# Flags
@break_on_fail_flag(help_message="Cancel upload of all files if one fails.")
@silent_flag(
    help_message="Turn off progress bar for each individual file. Summary bars still visible."
)
@click.pass_obj
def put_data(
    click_ctx,
    username,
    project,
    source,
    source_path_file,
    break_on_fail,
    overwrite,
    num_threads,
    silent,
):
    """Upload data to project.

    This first compressed the files (if not already compressed), encrypts them, and finally uploads
    them to Safespring S3 Storage.
    """
    try:
        dds_cli.data_putter.put(
            username=username,
            project=project,
            source=source,
            source_path_file=source_path_file,
            break_on_fail=break_on_fail,
            overwrite=overwrite,
            num_threads=num_threads,
            silent=silent,
            no_prompt=click_ctx.get("NO_PROMPT", False),
        )
    except (
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.UploadError,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds data get -- #
@data_group_command.command(name="get", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True, help_message="Project ID from which you're downloading data.")
@num_threads_option()
@source_option(help_message="Path to file or directory.", option_type=str)
@source_path_file_option()
@click.option(
    "--destination",
    "-d",
    required=False,
    type=click_pathlib.Path(exists=False, file_okay=False, dir_okay=True, resolve_path=True),
    multiple=False,
    help="Destination of downloaded files.",
)
# Flags
@break_on_fail_flag(help_message="Cancel download of all files if one fails.")
@silent_flag(
    help_message="Turn off progress bar for each individual file. Summary bars still visible."
)
@click.option(
    "--get-all",
    "-a",
    is_flag=True,
    default=False,
    show_default=True,
    help="Download all project contents.",
)
@click.option(
    "--verify-checksum",
    is_flag=True,
    default=False,
    show_default=True,
    help="Perform SHA-256 checksum verification after download (slower).",
)
@click.pass_obj
def get_data(
    click_ctx,
    username,
    project,
    get_all,
    source,
    source_path_file,
    destination,
    break_on_fail,
    num_threads,
    silent,
    verify_checksum,
):
    """Download files within a project.

    This downloads, decrypts, and finally decompresses (if compressed by the DDS) the files.
    """
    if get_all and (source or source_path_file):
        LOG.error(
            "Flag '--get-all' cannot be used together with options '--source'/'--source-path-fail'."
        )
        sys.exit(1)

    try:
        # Begin delivery
        with dds_cli.data_getter.DataGetter(
            username=username,
            project=project,
            get_all=get_all,
            source=source,
            source_path_file=source_path_file,
            break_on_fail=break_on_fail,
            destination=destination,
            silent=silent,
            verify_checksum=verify_checksum,
            no_prompt=click_ctx.get("NO_PROMPT", False),
        ) as getter:

            with rich.progress.Progress(
                "{task.description}",
                rich.progress.BarColumn(bar_width=None),
                " • ",
                "[progress.percentage]{task.percentage:>3.1f}%",
                refresh_per_second=2,
                console=dds_cli.utils.stderr_console,
            ) as progress:

                # Keep track of futures
                download_threads = {}

                # Iterator to keep track of which files have been handled
                iterator = iter(getter.filehandler.data.copy())

                with concurrent.futures.ThreadPoolExecutor() as texec:
                    task_dwnld = progress.add_task(
                        "Download", total=len(getter.filehandler.data), step="summary"
                    )

                    # Schedule the first num_threads futures for upload
                    for file in itertools.islice(iterator, num_threads):
                        LOG.debug(f"Starting: {file}")
                        # Execute download
                        download_threads[
                            texec.submit(getter.download_and_verify, file=file, progress=progress)
                        ] = file

                    while download_threads:
                        # Wait for the next future to complete
                        ddone, _ = concurrent.futures.wait(
                            download_threads, return_when=concurrent.futures.FIRST_COMPLETED
                        )

                        new_tasks = 0

                        for dfut in ddone:
                            downloaded_file = download_threads.pop(dfut)
                            LOG.debug(
                                f"Future done: {downloaded_file}",
                            )

                            # Get result
                            try:
                                file_downloaded = dfut.result()
                                LOG.debug(
                                    f"Download of {downloaded_file} successful: {file_downloaded}"
                                )
                            except concurrent.futures.BrokenExecutor as err:
                                LOG.critical(
                                    f"Download of file {downloaded_file} failed! Error: {err}"
                                )
                                continue

                            new_tasks += 1
                            progress.advance(task_dwnld)

                        # Schedule the next set of futures for download
                        for next_file in itertools.islice(iterator, new_tasks):
                            LOG.debug(f"Starting: {next_file}")
                            # Execute download
                            download_threads[
                                texec.submit(
                                    getter.download_and_verify,
                                    file=next_file,
                                    progress=progress,
                                )
                            ] = next_file
    except (
        dds_cli.exceptions.InvalidMethodError,
        OSError,
        dds_cli.exceptions.TokenNotFoundError,
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.ApiRequestError,
        dds_cli.exceptions.ApiResponseError,
        SystemExit,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.NoDataError,
        dds_cli.exceptions.DownloadError,
    ) as err:
        LOG.error(err)
        sys.exit(1)


# -- dds data ls -- #
@data_group_command.command(name="ls", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
@folder_option(help_message="List contents in this project folder.")
# Flags
@json_flag(help_message="Output in JSON format.")
@size_flag(help_message="Show size of project contents.")
@tree_flag(help_message="Display the entire project(s) directory tree.")
@users_flag(help_message="Display users associated with a project(Requires a project id).")
@click.pass_context
def list_data(ctx, username, project, folder, json, size, tree, users):
    """List project contents.

    Same as dds ls [PROJECT ID].
    """
    ctx.invoke(
        list_projects_and_contents,
        username=username,
        project=project,
        folder=folder,
        size=size,
        tree=tree,
        users=users,
        json=json,
    )


# -- dds data rm -- #
@data_group_command.command(name="rm", no_args_is_help=True)
# Options
@username_option()
@project_option(required=True)
@folder_option(
    help_message="Path to folder to remove.",
    short="-fl",
    multiple=True,
)
@click.option(
    "--file",
    "-f",
    required=False,
    type=str,
    multiple=True,
    help="Path to file to be removed." + dds_cli.utils.multiple_help_text(item="file"),
)
# Flags
@click.option(
    "--rm-all",
    "-a",
    is_flag=True,
    default=False,
    help="Remove all project contents.",
)
@click.pass_obj
def rm_data(click_ctx, username, project, file, folder, rm_all):
    """Delete project data."""
    no_prompt = click_ctx.get("NO_PROMPT", False)

    # Either all or a file
    if rm_all and (file or folder):
        LOG.error("The options '--rm-all' and '--file'/'--folder' cannot be used together.")
        sys.exit(1)

    # Will not delete anything if no file or folder specified
    if project and not any([rm_all, file, folder]):
        LOG.error(
            "One of the options must be specified to perform data deletion: "
            "'--rm-all' / '--file' / '--folder'."
        )
        sys.exit(1)

    # Warn if trying to remove all contents
    if rm_all:
        if no_prompt:
            LOG.warning(f"Deleting all files within project '{project}'")
        else:
            if not rich.prompt.Confirm.ask(
                f"Are you sure you want to delete all files within project '{project}'?"
            ):
                LOG.info("Probably for the best. Exiting.")
                sys.exit(0)

    try:
        with dds_cli.data_remover.DataRemover(
            project=project,
            username=username,
            no_prompt=no_prompt,
        ) as remover:

            if rm_all:
                remover.remove_all()

            elif file:
                remover.remove_file(files=file)

            elif folder:
                remover.remove_folder(folder=folder)
    except (
        dds_cli.exceptions.AuthenticationError,
        dds_cli.exceptions.APIError,
        dds_cli.exceptions.DDSCLIException,
        dds_cli.exceptions.ApiResponseError,
    ) as err:
        LOG.error(err)
        sys.exit(1)



import datetime

import pytest

from dds_cli.exceptions import TokenDeserializationError, TokenExpirationMissingError
from dds_cli.utils import get_token_expiration_time, get_token_header_contents, readable_timedelta

sample_fully_authenticated_token = (
    "eyJhbGciOiJBMjU2S1ciLCJlbmMiOiJBMjU2R0"
    "NNIiwiZXhwIjoiMjAyMi0wMi0yNFQxNDo1MTow"
    "Ni4yMTg3OTgifQ.Ul5rfhy0S9iaX2dPGH93HtL"
    "-3tVdGBdAzzoTQXb_QJrcIIA0wEwdQw.95ii5p"
    "anPoIUV1Mf.cAwnDuri4kxjwnQfY48pS0rZ-ob"
    "-RnKBacUcOe0l3RJrMCbc2nfkdkzc7KBH06ESi"
    "D-I7MU-U6270uLa2M4ZcLk0AkCZ3S7xrm9-bDu"
    "_73yCDCIQravwphlxCVSSrNQUPU8BonwBuDu-5"
    "WjuJyL_zC7MBcduxau8L0Hpk0IOLfIDgEtq9uR"
    "ELIxjbw1-YEhOtUBKm3E3jevmohgCt7RqcbbuB"
    "ZtZgYSm5NjOO1XhHBz_kZo1lhONNVVDNUkAoAP"
    "FoJ7WOAPajCGiDi8yyq7e-ojcxoSf0gl5NVd25"
    "cmO7i4OqsXB9VNlN5asEZE4WXAmVrQTppCbTG_"
    "9te04fCDwGabzDqtdfUqX-d_yaQ_UYHmJMN1xc"
    "4aF-uWZtk3loyMZU-uedQPqsJSZ.ay0MIzbtmt"
    "GsxUm2blaKUA"
)

token_without_exp_claim_in_header = (
    "eyJhbGciOiJBMjU2S1ciLCJlbmMiOiJBMjU2R"
    "0NNIn0.3H7fZh-rxkSuERSgknz4fOtseDn6PN"
    "c0RR-1IU8EmoTfpOuMOTvVbg.a_UwR9ArB6kn"
    "1LEB.7Ko4g1Xs9S_EQsAbGCtc96x_h3P6lwZz"
    "_X6t1EKA-EFeLXgwjAHuX5S_rC7YK28rqIT_m"
    "9FQABgTSgi0nBHCUurPA43U2P2mDR9UOvCHFY"
    "QXLKyO3M-ykVrmNwSGZMjo3HHrmcuICiwiH7l"
    "boGl5Vr-iSFpyyuy33thSrlwfutI80sKe3RSm"
    "Kup_Mh7tM0mw0WbQezfAcNR_52BeP_ncbVxFl"
    "714ikyo2HCk0bKREIpetdaKCaoZgqlhOarlAU"
    "GwPaKtdgmXb7Ef4VKfYdnLIxqzv3RtVmZiEb1"
    "L-xCS4vnwXvw_bEa_QU-5HfyLYOszjAHiYHxr"
    "q8v1xnfoyWfd20OxQMhYueVzlPw1HfMSfvCNV"
    "LZO-vNNKHTaGnPyGuykhMNScIgkR1l8.TEp4L"
    "s4c29JtGogmdYbTbw"
)


def test_get_token_expiration_time_successful():
    exp_claim_in_token_header = get_token_expiration_time(token=sample_fully_authenticated_token)
    assert isinstance(datetime.datetime.fromisoformat(exp_claim_in_token_header), datetime.datetime)


def test_get_token_expiration_time_exception():
    with pytest.raises(TokenExpirationMissingError) as error:
        get_token_expiration_time(token=token_without_exp_claim_in_header)

    assert "Expiration time could not be found in the header of the token." in str(error.value)


def test_get_token_header_contents_exception():
    with pytest.raises(TokenDeserializationError) as error:
        get_token_header_contents(token="not.a.token")

    assert "Token could not be deserialized." in str(error.value)

    with pytest.raises(TokenDeserializationError) as error:
        get_token_header_contents(token="notatoken")

    assert "Token could not be deserialized." in str(error.value)

    with pytest.raises(TokenDeserializationError) as error:
        get_token_header_contents(token="not.a.token.not.a")

    assert "Token could not be deserialized." in str(error.value)


def test_readable_timedelta():
    # at the time of writing this test, negative values are not passed in, but the function already supports it
    assert readable_timedelta(datetime.timedelta(seconds=-60)) == "1 minute"
    assert readable_timedelta(datetime.timedelta(milliseconds=-100)) == "less than a minute"

    assert readable_timedelta(datetime.timedelta(milliseconds=100)) == "less than a minute"
    assert readable_timedelta(datetime.timedelta(seconds=59)) == "less than a minute"
    assert readable_timedelta(datetime.timedelta(seconds=60)) == "1 minute"
    assert readable_timedelta(datetime.timedelta(minutes=1)) == "1 minute"
    assert readable_timedelta(datetime.timedelta(seconds=98765)) == "1 day 3 hours 26 minutes"
    assert readable_timedelta(datetime.timedelta(hours=3)) == "3 hours"
    assert readable_timedelta(datetime.timedelta(days=1)) == "1 day"



"""Data Delivery System Project Status manager."""
import logging

# Installed
import requests
import simplejson
import pytz
import tzlocal
import datetime

# Own modules
from dds_cli import base
from dds_cli import exceptions
from dds_cli import DDSEndpoint
import dds_cli.utils

###############################################################################
# START LOGGING CONFIG ################################# START LOGGING CONFIG #
###############################################################################

LOG = logging.getLogger(__name__)


###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


class ProjectStatusManager(base.DDSBaseClass):
    """Project Status manager class."""

    def __init__(
        self,
        username: str,
        project: str,
        no_prompt: bool = False,
    ):
        """Handle actions regarding project status in the cli."""
        # Initiate DDSBaseClass to authenticate user
        super().__init__(
            username=username,
            no_prompt=no_prompt,
            method_check=False,
        )
        self.project = project

    # Public methods ###################### Public methods #
    def get_status(self, show_history):
        """Get current status and status history of the project"""

        try:
            response = requests.get(
                DDSEndpoint.UPDATE_PROJ_STATUS,
                headers=self.token,
                params={"project": self.project},
                json={"history": show_history},
            )
        except requests.exceptions.RequestException as err:
            raise exceptions.ApiRequestError(message=str(err))

        # Check response
        if not response.ok:
            raise exceptions.APIError(f"Failed to get any projects: {response.text}")

        # Get result from API
        try:
            resp_json = response.json()
        except simplejson.JSONDecodeError as err:
            raise exceptions.APIError(f"Could not decode JSON response: {err}")
        else:
            current_status = resp_json.get("current_status")
            current_deadline = resp_json.get("current_deadline")
            status_out = f"Current status of {self.project}: {current_status}"
            deadline_out = ""
            if current_deadline:
                try:
                    date = pytz.timezone("UTC").localize(
                        datetime.datetime.strptime(current_deadline, "%a, %d %b %Y %H:%M:%S GMT")
                    )
                except ValueError as err:
                    raise exceptions.ApiResponseError(
                        f"Time zone mismatch: Incorrect zone '{current_deadline.split()[-1]}'"
                    )
                else:
                    current_deadline = date.astimezone(tzlocal.get_localzone()).strftime(
                        "%a, %d %b %Y %H:%M:%S %Z"
                    )
                deadline_out = f" with deadline {current_deadline}"
            dds_cli.utils.console.print(f"{status_out}{deadline_out}")
            if show_history:
                history = "Status history \n"
                for row in resp_json.get("history"):
                    try:
                        date = pytz.timezone("UTC").localize(
                            datetime.datetime.strptime(row[1], "%a, %d %b %Y %H:%M:%S GMT")
                        )
                    except ValueError as err:
                        raise exceptions.ApiResponseError(
                            f"Time zone mismatch: Incorrect zone '{row[1].split()[-1]}'"
                        )
                    else:
                        row[1] = date.astimezone(tzlocal.get_localzone()).strftime(
                            "%a, %d %b %Y %H:%M:%S %Z"
                        )
                    history += ", ".join([item for item in row]) + " \n"
                LOG.info(history)

    def update_status(self, new_status, deadline=None, is_aborted=False, no_mail=False):
        """Update project status"""

        extra_params = {"new_status": new_status, "send_email": not no_mail}
        if deadline:
            extra_params["deadline"] = deadline
        if is_aborted:
            extra_params["is_aborted"] = is_aborted
        try:
            response = requests.post(
                DDSEndpoint.UPDATE_PROJ_STATUS,
                headers=self.token,
                params={"project": self.project},
                json=extra_params,
            )
        except requests.exceptions.RequestException as err:
            raise exceptions.ApiRequestError(message=str(err))

        # Check response
        if not response.ok:
            raise exceptions.APIError(f"An Error occured: {response.json().get('message')}")

        # Get result from API
        try:
            resp_json = response.json()
        except simplejson.JSONDecodeError as err:
            raise exceptions.APIError(f"Could not decode JSON response: {err}")
        else:
            dds_cli.utils.console.print(f"Project {resp_json.get('message')}")



# Standard library
import unittest.mock

# Installed
import pytest
import requests

# Own modules
import dds_cli

ADD_JSON = {"email": "test.testsson@example.com", "role": "Researcher"}
PROJECT = {"project": "test_project"}


@pytest.fixture
def runner_with_project(runner):
    """Run dds add-user with a project specified."""

    def _run():
        return runner(
            [
                "user",
                "add",
                "-u",
                "unituser",
                "-r",
                ADD_JSON["role"],
                "-p",
                PROJECT["project"],
                "--no-mail",
                ADD_JSON["email"],
            ]
        )

    yield _run


@pytest.fixture
def runner_no_project(runner):
    """Run dds add-user without a project specified."""

    def _run():
        return runner(["user", "add", "-u", "unituser", "-r", ADD_JSON["role"], ADD_JSON["email"]])

    yield _run


@pytest.fixture
def add_user():
    """A fixture that mocks the requests.post method.

    The functioned returned by this fixture takes parameters that adjust the status_code,
    ok, and json message.
    """
    with unittest.mock.patch.object(requests, "post") as mock_B:

        def _request_mock(status_code, message=None, ok=True):
            mock_returned_request = unittest.mock.MagicMock(status_code=status_code, ok=ok)
            mock_returned_request.json.return_value = {"message": message}
            mock_B.return_value = mock_returned_request
            return mock_B

        yield _request_mock


def test_add_user_no_project_OK(runner_no_project, add_user):
    add_user_OK = add_user(200)
    result = runner_no_project()
    add_user_OK.assert_called_with(
        dds_cli.DDSEndpoint.USER_ADD,
        json={**ADD_JSON, "send_email": True},
        params={"project": None},
        headers=unittest.mock.ANY,
    )

    assert result.exit_code == 0


def test_add_user_no_project_fail(runner_no_project, add_user):
    add_user_FAIL = add_user(403, message="Specifically passed message", ok=False)
    result = runner_no_project()
    add_user_FAIL.assert_called_with(
        dds_cli.DDSEndpoint.USER_ADD,
        json={**ADD_JSON, "send_email": True},
        params={"project": None},
        headers=unittest.mock.ANY,
    )

    assert "Could not add user" in result.stderr
    assert "Specifically passed message" in result.stderr
    assert result.exit_code != 0


def test_add_user_with_project_ok(runner_with_project, add_user):
    add_user_OK = add_user(200)
    result = runner_with_project()
    add_user_OK.assert_called_with(
        dds_cli.DDSEndpoint.USER_ADD,
        json={**ADD_JSON, "send_email": False},
        params=PROJECT,
        headers=unittest.mock.ANY,
    )

    assert result.exit_code == 0


def test_add_user_with_project_fail(runner_with_project, add_user):
    add_user_OK = add_user(403, message="Specifically passed message", ok=False)
    result = runner_with_project()
    add_user_OK.assert_called_with(
        dds_cli.DDSEndpoint.USER_ADD,
        json={**ADD_JSON, "send_email": False},
        params=PROJECT,
        headers=unittest.mock.ANY,
    )

    assert "Could not add user" in result.stderr
    assert "Specifically passed message" in result.stderr
    assert result.exit_code != 0



"""Data putter."""

###############################################################################
# IMPORTS ########################################################### IMPORTS #
###############################################################################

# Standard library
import concurrent.futures
import itertools
import logging
import pathlib

# Installed
import boto3
import botocore
import requests
from rich.progress import Progress, SpinnerColumn, BarColumn
import simplejson

# Own modules
from dds_cli import base
from dds_cli import exceptions
from dds_cli import data_remover as dr
from dds_cli import DDSEndpoint
from dds_cli import file_encryptor as fe
from dds_cli import file_handler_local as fhl
from dds_cli import status
from dds_cli import text_handler as txt
from dds_cli.custom_decorators import verify_proceed, update_status, subpath_required

import dds_cli
import dds_cli.utils

###############################################################################
# START LOGGING CONFIG ################################# START LOGGING CONFIG #
###############################################################################

LOG = logging.getLogger(__name__)

###############################################################################
# MAIN FUNCTION ############################################### MAIN FUNCTION #
###############################################################################


def put(
    username,
    project,
    source,
    source_path_file,
    break_on_fail,
    overwrite,
    num_threads,
    silent,
    no_prompt,
):
    """Handle upload of data."""
    # Initialize delivery - check user access etc
    with DataPutter(
        username=username,
        project=project,
        source=source,
        source_path_file=source_path_file,
        break_on_fail=break_on_fail,
        overwrite=overwrite,
        silent=silent,
        no_prompt=no_prompt,
    ) as putter:

        # Progress object to keep track of progress tasks
        with Progress(
            "{task.description}",
            BarColumn(bar_width=None),
            " • ",
            "[progress.percentage]{task.percentage:>3.1f}%",
            refresh_per_second=2,
            console=dds_cli.utils.stderr_console,
        ) as progress:

            # Keep track of futures
            upload_threads = {}

            # Iterator to keep track of which files have been handled
            iterator = iter(putter.filehandler.data.copy())

            with concurrent.futures.ThreadPoolExecutor() as texec:
                # Start main progress bar - total uploaded files
                upload_task = progress.add_task(
                    description="Upload",
                    total=len(putter.filehandler.data),
                )

                # Schedule the first num_threads futures for upload
                for file in itertools.islice(iterator, num_threads):
                    LOG.debug(f"Starting: {file}")
                    upload_threads[
                        texec.submit(
                            putter.protect_and_upload,
                            file=file,
                            progress=progress,
                        )
                    ] = file

                try:
                    # Continue until all files are done
                    while upload_threads:
                        # Wait for the next future to complete, _ are the unfinished
                        done, _ = concurrent.futures.wait(
                            upload_threads,
                            return_when=concurrent.futures.FIRST_COMPLETED,
                        )

                        # Number of new upload tasks that can be started
                        new_tasks = 0

                        # Get result from future and schedule database update
                        for fut in done:
                            uploaded_file = upload_threads.pop(fut)
                            LOG.debug(f"Future done for file: {uploaded_file}")

                            # Get result
                            try:
                                file_uploaded = fut.result()
                                LOG.debug(f"Upload of {uploaded_file} successful: {file_uploaded}")
                            except concurrent.futures.BrokenExecutor as err:
                                LOG.error(f"Upload of file {uploaded_file} failed! Error: {err}")
                                continue

                            # Increase the main progress bar
                            progress.advance(upload_task)

                            # New available threads
                            new_tasks += 1

                        # Schedule the next set of futures for upload
                        for next_file in itertools.islice(iterator, new_tasks):
                            LOG.debug(f"Starting: {next_file}")
                            upload_threads[
                                texec.submit(
                                    putter.protect_and_upload,
                                    file=next_file,
                                    progress=progress,
                                )
                            ] = next_file
                except KeyboardInterrupt:
                    LOG.warning(
                        "KeyboardInterrupt found - shutting down delivery gracefully. "
                        "This will finish the ongoing uploads. If you want to force "
                        "shutdown, repeat `Ctrl+C`. This is not advised. "
                    )

                    # Flag for threads to find
                    putter.stop_doing = True

                    # Stop and remove main progress bar
                    progress.remove_task(upload_task)

                    # Stop all tasks that are not currently uploading
                    _ = [
                        progress.stop_task(x)
                        for x in [y.id for y in progress.tasks if y.fields.get("step") != "put"]
                    ]


###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


class DataPutter(base.DDSBaseClass):
    """Data putter class."""

    def __init__(
        self,
        username: str,
        project: str = None,
        break_on_fail: bool = False,
        overwrite: bool = False,
        source: tuple = (),
        source_path_file: pathlib.Path = None,
        silent: bool = False,
        method: str = "put",
        no_prompt: bool = False,
    ):
        """Handle actions regarding upload of data."""
        # Initiate DDSBaseClass to authenticate user
        super().__init__(username=username, project=project, method=method, no_prompt=no_prompt)

        # Initiate DataPutter specific attributes
        self.break_on_fail = break_on_fail
        self.overwrite = overwrite
        self.silent = silent
        self.filehandler = None

        # Only method "put" can use the DataPutter class
        if self.method != "put":
            raise exceptions.AuthenticationError(f"Unauthorized method: '{self.method}'")

        # Start file prep progress
        with Progress(
            "[bold]{task.description}",
            SpinnerColumn(spinner_name="dots12", style="white"),
            console=dds_cli.utils.stderr_console,
        ) as progress:
            # Spinner while collecting file info
            wait_task = progress.add_task("Collecting and preparing data", step="prepare")

            # Get file info
            self.filehandler = fhl.LocalFileHandler(
                user_input=(source, source_path_file),
                project=self.project,
                temporary_destination=self.dds_directory.directories["FILES"],
            )

            # Verify that the Safespring S3 bucket exists
            # self.verify_bucket_exist()

            # Check which, if any, files exist in the db
            files_in_db = self.filehandler.check_previous_upload(token=self.token)

            # Quit if error and flag
            if files_in_db and self.break_on_fail and not self.overwrite:
                raise exceptions.UploadError(
                    "Some files have already been uploaded (or have identical names to "
                    "previously uploaded files) and the '--break-on-fail' flag was used. "
                    "Try again with the '--overwrite' flag if you want to upload these files."
                )

            # Generate status dict
            self.status = self.filehandler.create_upload_status_dict(
                existing_files=files_in_db, overwrite=self.overwrite
            )

            # Remove spinner
            progress.remove_task(wait_task)

        if not self.filehandler.data:
            raise exceptions.UploadError("No data to upload.")

    # Public methods ###################### Public methods #
    @verify_proceed
    @subpath_required
    def protect_and_upload(self, file, progress):
        """Process and upload the file while handling the progress bars."""
        # Variables
        all_ok, saved, message = (False, False, "")  # Error catching
        file_info = self.filehandler.data[file]  # Info on current file
        file_public_key, salt = ("", "")  # Crypto info

        # Progress bar for processing
        task = progress.add_task(
            description=txt.TextHandler.task_name(file=file, step="encrypt"),
            total=file_info["size_raw"],
            visible=not self.silent,
        )

        # Stream chunks from file
        streamed_chunks = self.filehandler.stream_from_file(file=file)

        # Stream the chunks into the encryptor to save the encrypted chunks
        with fe.Encryptor(project_keys=self.keys) as encryptor:

            # Encrypt and save chunks
            saved, message = encryptor.encrypt_filechunks(
                chunks=streamed_chunks,
                outfile=file_info["path_processed"],
                progress=(progress, task),
            )

            # Get hex version of public key -- saved in db
            file_public_key = encryptor.get_public_component_hex(private_key=encryptor.my_private)
            salt = encryptor.salt

        LOG.debug(f"Updating file processed size: {file_info['path_processed']}")

        # Update file info incl size, public key, salt
        self.filehandler.data[file]["public_key"] = file_public_key
        self.filehandler.data[file]["salt"] = salt
        self.filehandler.data[file]["size_processed"] = file_info["path_processed"].stat().st_size

        if saved:
            LOG.debug(
                f"File successfully encrypted: {file}. New location: {file_info['path_processed']}"
            )
            # Update progress bar for upload
            progress.reset(
                task,
                description=txt.TextHandler.task_name(file=file, step="put"),
                total=self.filehandler.data[file]["size_processed"],
                step="put",
            )

            # Perform upload
            file_uploaded, message = self.put(file=file, progress=progress, task=task)

            # Perform db update
            if file_uploaded:
                db_updated, message = self.add_file_db(file=file)

                if db_updated:
                    all_ok = True
                    LOG.debug(f"File successfully uploaded and added to the database: {file}")

        if not saved or all_ok:
            # Delete temporary processed file locally
            LOG.debug(
                f"Deleting file {file_info['path_processed']} - "
                f"exists: {file_info['path_processed'].exists()}"
            )
            dr.DataRemover.delete_tempfile(file=file_info["path_processed"])

        # Remove progress bar task
        progress.remove_task(task)

        return all_ok, message

    @update_status
    def put(self, file, progress, task):
        """Upload files to the cloud."""
        # Variables
        uploaded = False
        error = ""

        # File info
        file_local = str(self.filehandler.data[file]["path_processed"])
        file_remote = self.filehandler.data[file]["path_remote"]

        try:
            with self.s3connector as conn:
                # Upload file
                conn.resource.meta.client.upload_file(
                    Filename=file_local,
                    Bucket=conn.bucketname,
                    Key=file_remote,
                    ExtraArgs={
                        "ACL": "private",  # Access control list
                        "CacheControl": "no-store",  # Don't store cache
                    },
                    Callback=status.ProgressPercentage(
                        progress=progress,
                        task=task,
                    )
                    if task is not None
                    else None,
                )
        except (
            botocore.client.ClientError,
            boto3.exceptions.Boto3Error,
            botocore.exceptions.BotoCoreError,
            FileNotFoundError,
            TypeError,
        ) as err:
            error = f"S3 upload of file '{file}' failed: {err}"
            LOG.exception(f"{file}: {err}")
        else:
            uploaded = True

        return uploaded, error

    @update_status
    def add_file_db(self, file):
        """Make API request to add file to DB."""
        # Variables
        added_to_db = False
        error = ""

        # Get file info and specify info required in db
        fileinfo = self.filehandler.data[file]
        LOG.debug(f"Fileinfo: {fileinfo}")
        params = {"project": self.project}
        file_info = {
            "name": file,
            "name_in_bucket": str(fileinfo["path_remote"]),
            "subpath": str(fileinfo["subpath"]),
            "size": fileinfo["size_raw"],
            "size_processed": fileinfo["size_processed"],
            "compressed": not fileinfo["compressed"],
            "salt": fileinfo["salt"],
            "public_key": fileinfo["public_key"],
            "checksum": fileinfo["checksum"],
        }

        # Send file info to API - post if new file, put if overwrite
        put_or_post = requests.put if fileinfo["overwrite"] else requests.post
        try:
            response = put_or_post(
                DDSEndpoint.FILE_NEW,
                params=params,
                json=file_info,
                headers=self.token,
                timeout=DDSEndpoint.TIMEOUT,
            )
        except requests.exceptions.RequestException as err:
            error = str(err)
            LOG.warning(error)
        else:
            # Error if failed
            if not response.ok:
                error = f"Failed to add file '{file}' to database: {response.text}"
                return added_to_db, error

            try:
                added_to_db, error = (True, response.json().get("message"))
            except simplejson.JSONDecodeError as err:
                error = str(err)
                LOG.warning(error)

        return added_to_db, error



"""File handler module. Base class for LocalFileHandler and RemoteFileHandler."""

###############################################################################
# IMPORTS ########################################################### IMPORTS #
###############################################################################

# Standard library
import json
import logging
import os
import pathlib

# Installed

# Own modules
import dds_cli.utils
import dds_cli.exceptions

###############################################################################
# START LOGGING CONFIG ################################# START LOGGING CONFIG #
###############################################################################

LOG = logging.getLogger(__name__)

###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


class FileHandler:
    """Main file handler."""

    def __init__(self, user_input, local_destination, project=None):
        """Initiate file handler."""
        source, source_path_file = user_input

        # Get user specified data
        self.project = project
        self.local_destination = local_destination
        self.data_list = []
        if source is not None:
            self.data_list += list(source)
        if source_path_file is not None:
            source_path_file = pathlib.Path(source_path_file)
            if source_path_file.exists():
                try:
                    with source_path_file.resolve().open(mode="r") as spf:
                        self.data_list += spf.read().splitlines()
                except OSError as err:
                    dds_cli.utils.console.print(
                        f"Failed to get files from source-path-file option: {err}"
                    )
                    os._exit(1)

        self.failed = {}

    # Static methods ############ Static methods #
    @staticmethod
    def append_errors_to_file(log_file: pathlib.Path, file, info, status):
        """Save errors to specific json file."""

        failed_to_save = {
            str(file): {
                **FileHandler.make_json_serializable(non_json=info),
                "status": FileHandler.make_json_serializable(non_json=status),
            }
        }
        try:
            with log_file.open(mode="a") as errfile:
                json_output = json.dumps(
                    failed_to_save,
                    indent=4,
                )
                # Each line is valid json, but the entire file is not.
                # Multiple threads are appending to this file, so valid json for
                # the entire file is not trivial.
                errfile.write(json_output + "\n")
        except (OSError, TypeError) as err:
            LOG.warning(str(err))

    @staticmethod
    def make_json_serializable(non_json):
        """Convert pathlib.Path instances in dict to string."""
        return {str(x): (str(y) if isinstance(y, pathlib.Path) else y) for x, y in non_json.items()}

    @staticmethod
    def delete_tempdir(directory: pathlib.Path):
        """Delete the specified directory."""
        ok_to_remove = False

        # If file not ok to remove folder
        if not directory.is_dir():
            return ok_to_remove

        # Iterate through any existing subdirectories - recursive
        LOG.debug(f"Any in directory? {any(directory.iterdir())}")
        for x in directory.iterdir():
            LOG.debug(x)
        if any(directory.iterdir()):
            for p in directory.iterdir():
                if p.is_dir():
                    ok_to_remove = FileHandler.delete_tempdir(directory=p)
                    if ok_to_remove:
                        directory.rmdir()
        else:
            directory.rmdir()
            ok_to_remove = True

        return ok_to_remove



"""File encryptor module. Handles the encryption of the files and the generation of the keys."""

###############################################################################
# IMPORTS ########################################################### IMPORTS #
###############################################################################

# Standard library
import hashlib
import logging
import os
import pathlib
import traceback

# Installed
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives.kdf import hkdf
from nacl.bindings import crypto_aead_chacha20poly1305_ietf_decrypt
from nacl.bindings import crypto_aead_chacha20poly1305_ietf_encrypt

# Own modules
from dds_cli import FileSegment
from dds_cli.file_handler_local import LocalFileHandler as fh

###############################################################################
# START LOGGING CONFIG ################################# START LOGGING CONFIG #
###############################################################################

LOG = logging.getLogger(__name__)


###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


class ECDHKeyHandler:
    """Generates the shared encryption/decryption key, and transforms the components."""

    # Static methods ###################### Statis methods #
    @staticmethod
    def generate_shared_key(my_private, peer_public, salt: bytes = b""):
        """Derive the shared key for file encryption."""

        # Generate or from db
        if salt == b"":
            salt = os.urandom(16)

        # Project public key
        # peer_public_bytes = bytes.fromhex(peer_public)
        # loaded_peer_public = x25519.X25519PublicKey.from_public_bytes(peer_public_bytes)

        # Generate shared key and derive encryption key with salt
        shared_key = (my_private).exchange(peer_public_key=peer_public)
        derived_shared_key = hkdf.HKDF(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            info=b"",
            backend=backends.default_backend(),
        ).derive(shared_key)

        LOG.debug(f"Salt: {salt}")
        return derived_shared_key, salt.hex().upper()

    @staticmethod
    def public_to_hex(public_key):
        """Converts public key to hexstring."""

        # public = self.private.public_key()
        public_bytes = public_key.public_bytes(
            encoding=serialization.Encoding.Raw, format=serialization.PublicFormat.Raw
        )

        return public_bytes.hex().upper()

    @staticmethod
    def get_public_component_hex(private_key):
        """Gets the public key and converts to hex string."""

        public = private_key.public_key()
        public_bytes = public.public_bytes(
            encoding=serialization.Encoding.Raw, format=serialization.PublicFormat.Raw
        )

        return public_bytes.hex().upper()


class Encryptor(ECDHKeyHandler):
    """Handles the encryption of the files."""

    def __init__(self, project_keys):
        self.max_nonce = 2 ** (12 * 8)  # Max mumber of nonces

        # Only peer public needed, private should be None
        self.peer_public = x25519.X25519PublicKey.from_public_bytes(bytes.fromhex(project_keys[1]))

        # This generates public too
        self.my_private = x25519.X25519PrivateKey.generate()

        self.key, self.salt = self.generate_shared_key(
            my_private=self.my_private, peer_public=self.peer_public
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
            return False  # uncomment to pass exception through

        return True

    # Static methods ###################### Static methods #
    @staticmethod
    def verify_checksum(file: pathlib.Path, correct_checksum):
        """Generate file checksum and verify the integrity"""

        verified, error = (False, "")

        checksum = hashlib.sha256()

        try:
            for chunk in fh.read_file(file=file):
                checksum.update(chunk)
        except OSError as err:
            error = str(err)
        else:
            if checksum.hexdigest() == correct_checksum:
                verified, error = (True, "File integrity verified.")
                LOG.info("Checksum verification successful. File integrity verified.")
                LOG.debug(error)
            else:
                error = "Checksum verification failed. File compromised."
                LOG.warning(error)

        return verified, error

    # Public methods ###################### Public methods #
    def encrypt_filechunks(self, chunks, outfile: pathlib.Path, progress: tuple = None):
        """Encrypts the file in chunks.

        Encrypts the file in chunks using the IETF ratified ChaCha20-Poly1305
        construction described in RFC8439 (obsoletes 7539).
        """

        LOG.debug("Started encryption...")
        encrypted_and_saved, message = (False, "")

        # Additional data
        aad = None

        try:
            # Save encryption output to file
            with outfile.open(mode="wb") as out:
                # Create and save first IV/nonce
                iv_bytes = os.urandom(12)
                out.write(iv_bytes)

                # Get first iv/nonce as integer
                iv_int = int.from_bytes(iv_bytes, "little")
                nonce = b""  # Catch last nonce
                for chunk in chunks:

                    # Restart at 0 if nonce number at maximum number of chunks per key
                    nonce = (
                        iv_int if iv_int < self.max_nonce else iv_int % self.max_nonce
                    ).to_bytes(length=12, byteorder="little")

                    # Encrypt chunk
                    encrypted_chunk = crypto_aead_chacha20poly1305_ietf_encrypt(
                        message=chunk, aad=aad, nonce=nonce, key=self.key
                    )
                    out.write(encrypted_chunk)

                    progress[0].advance(progress[1], FileSegment.SEGMENT_SIZE_RAW)
                    iv_int += 1  # Increment nonce

                # Save last nonce
                out.write(nonce)
        except (OSError, TypeError, FileExistsError, InterruptedError) as err:
            message = str(err)
            LOG.exception(message)
        else:
            encrypted_and_saved = True
            message = f"Encrypted file stored in location: {outfile}"
            LOG.debug(message)

        return encrypted_and_saved, message


class Decryptor(ECDHKeyHandler):
    """Handles the decryption of the files."""

    def __init__(self, project_keys: tuple, peer_public: str, key_salt: str):

        self.max_nonce = 2 ** (12 * 8)

        # Only private needed, public generated from it.
        self.my_private = x25519.X25519PrivateKey.from_private_bytes(bytes.fromhex(project_keys[0]))

        # Only peer public used
        self.peer_public = x25519.X25519PublicKey.from_public_bytes(bytes.fromhex(peer_public))

        self.key, _ = self.generate_shared_key(
            my_private=self.my_private,
            peer_public=self.peer_public,
            salt=bytes.fromhex(key_salt),
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            traceback.print_exception(exc_type, exc_value, tb)
            return False  # uncomment to pass exception through

        return True

    # Public methods ###################### Public methods #
    def decrypt_file(self, infile: pathlib.Path):
        """Decrypts the file"""

        try:
            with infile.open(mode="rb+") as file:
                # Get last nonce
                file.seek(-12, os.SEEK_END)
                last_nonce = file.read(12)

                # Remove last nonce from file
                file.seek(-12, os.SEEK_END)
                file.truncate()

                # Jump back to beginning and get first nonce
                file.seek(0)
                first_nonce = file.read(12)

                # Decrypt file
                if file.tell() != 12:
                    raise SystemExit

                iv_int = int.from_bytes(first_nonce, "little")
                aad = None
                nonce = b""

                for chunk in iter(lambda: file.read(FileSegment.SEGMENT_SIZE_CIPHER), b""):
                    # Get nonce as bytes for decryption: if the nonce is larger than the
                    # max number of chunks allowed - wrap to 0 again
                    nonce = (
                        iv_int if iv_int < self.max_nonce else iv_int % self.max_nonce
                    ).to_bytes(length=12, byteorder="little")

                    iv_int += 1

                    yield crypto_aead_chacha20poly1305_ietf_decrypt(
                        ciphertext=chunk, aad=aad, nonce=nonce, key=self.key
                    )

                LOG.debug("Testing nonce...")
                if last_nonce != nonce:
                    raise SystemExit("Nonces do not match!!")
                LOG.debug(f"Last nonce should be: {last_nonce}, was: {nonce}")
        except Exception as err:
            LOG.warning(str(err))



"""Data getter."""

###############################################################################
# IMPORTS ########################################################### IMPORTS #
###############################################################################

# Standard library
import logging
import pathlib

# Installed
import requests
import simplejson
from rich.progress import Progress, SpinnerColumn

# Own modules
from dds_cli import DDSEndpoint, FileSegment
from dds_cli import file_handler_remote as fhr
from dds_cli import data_remover as dr
from dds_cli import file_compressor as fc
from dds_cli import file_encryptor as fe
from dds_cli import text_handler as txt
from dds_cli.custom_decorators import verify_proceed, update_status, subpath_required
from dds_cli import base
import dds_cli.utils

###############################################################################
# START LOGGING CONFIG ################################# START LOGGING CONFIG #
###############################################################################

LOG = logging.getLogger(__name__)

###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


class DataGetter(base.DDSBaseClass):
    """Data getter class."""

    def __init__(
        self,
        username: str,
        project: str = None,
        break_on_fail: bool = False,
        get_all: bool = False,
        source: tuple = (),
        source_path_file: pathlib.Path = None,
        destination: pathlib.Path = pathlib.Path(""),
        silent: bool = False,
        verify_checksum: bool = False,
        method: str = "get",
        no_prompt: bool = False,
    ):
        """Handle actions regarding downloading data."""
        # Initiate DDSBaseClass to authenticate user
        super().__init__(
            username=username,
            project=project,
            dds_directory=destination,
            method=method,
            no_prompt=no_prompt,
        )

        # Initiate DataGetter specific attributes
        self.break_on_fail = break_on_fail
        self.verify_checksum = verify_checksum
        self.silent = silent
        self.filehandler = None

        # Only method "get" can use the DataGetter class
        if self.method != "get":
            raise dds_cli.exceptions.InvalidMethodError(
                attempted_method=self.method,
                message="DataGetter attempting unauthorized method",
            )

        # Start file prep progress
        with Progress(
            "[bold]{task.description}",
            SpinnerColumn(spinner_name="dots12", style="white"),
            console=dds_cli.utils.stderr_console,
        ) as progress:
            wait_task = progress.add_task("Collecting and preparing data", step="prepare")
            self.filehandler = fhr.RemoteFileHandler(
                get_all=get_all,
                user_input=(source, source_path_file),
                token=self.token,
                project=self.project,
                destination=self.dds_directory.directories["FILES"],
            )

            if self.filehandler.failed and self.break_on_fail:
                raise dds_cli.exceptions.DownloadError(
                    ":warning-emoji: Some specified files were not found in the system "
                    "and '--break-on-fail' flag used. :warning-emoji:"
                    f"Files not found: {self.filehandler.failed}"
                )

            if not self.filehandler.data:
                raise dds_cli.exceptions.DownloadError("No files to download.")

            self.status = self.filehandler.create_download_status_dict()

            progress.remove_task(wait_task)

    # Public methods ############ Public methods #
    @verify_proceed
    @subpath_required
    def download_and_verify(self, file, progress):
        """Download the file, reveals the original data and verifies the integrity."""
        all_ok, message = (False, "")
        file_info = self.filehandler.data[file]

        # File task for downloading
        task = progress.add_task(
            description=txt.TextHandler.task_name(file=file, step="get"),
            total=file_info["size_stored"],
            visible=not self.silent,
        )

        # Perform download
        file_downloaded, message = self.get(file=file, progress=progress, task=task)

        # Update progress task for decryption
        progress.reset(
            task,
            description=txt.TextHandler.task_name(file=file, step="decrypt"),
            total=file_info["size_original"],
        )

        LOG.debug(f"File {file} downloaded: {file_downloaded}")

        if file_downloaded:
            db_updated, message = self.update_db(file=file)
            LOG.debug(f"Database updated: {db_updated}")

            LOG.debug(f"Beginning decryption of file {file}...")
            file_saved = False
            with fe.Decryptor(
                project_keys=self.keys,
                peer_public=file_info["public_key"],
                key_salt=file_info["salt"],
            ) as decryptor:

                streamed_chunks = decryptor.decrypt_file(infile=file_info["path_downloaded"])

                stream_to_file_func = (
                    fc.Compressor.decompress_filechunks
                    if file_info["compressed"]
                    else self.filehandler.write_file
                )

                file_saved, message = stream_to_file_func(
                    chunks=streamed_chunks,
                    outfile=file,
                )

            LOG.debug(f"file saved? {file_saved}")
            if file_saved:
                # TODO (ina): decide on checksum verification method --
                # this checks original, the other is generated from compressed
                all_ok, message = (
                    fe.Encryptor.verify_checksum(file=file, correct_checksum=file_info["checksum"])
                    if self.verify_checksum
                    else (True, "")
                )

            dr.DataRemover.delete_tempfile(file=file_info["path_downloaded"])

        progress.remove_task(task)
        return all_ok, message

    @update_status
    def get(self, file, progress, task):
        """Download files from the cloud."""
        downloaded = False
        error = ""
        file_local = self.filehandler.data[file]["path_downloaded"]
        file_remote = self.filehandler.data[file]["url"]

        try:
            with requests.get(file_remote, stream=True) as req:
                req.raise_for_status()
                with file_local.open(mode="wb") as new_file:
                    for chunk in req.iter_content(chunk_size=FileSegment.SEGMENT_SIZE_CIPHER):
                        progress.update(task, advance=len(chunk))
                        new_file.write(chunk)
        except (
            requests.exceptions.ConnectTimeout,
            requests.exceptions.HTTPError,
            requests.exceptions.ReadTimeout,
            requests.exceptions.Timeout,
            requests.exceptions.ConnectionError,
        ) as err:
            if (
                hasattr(err, "response")
                and hasattr(err.response, "status_code")
                and err.response.status_code == 404
            ):
                error = "File not found! Please report this to the SciLifeLab Data Centre."
            else:
                error = str(err)
        else:
            downloaded = True

        return downloaded, error

    @update_status
    def update_db(self, file):
        """Update file info in db."""
        updated_in_db = False
        error = ""

        # Get file info
        fileinfo = self.filehandler.data[file]
        filename = {"name": fileinfo["name_in_db"]}
        params = {"project": self.project}

        # Send file info to API
        try:
            response = requests.put(
                DDSEndpoint.FILE_UPDATE, params=params, json=filename, headers=self.token
            )
        except requests.exceptions.RequestException as err:
            raise SystemExit from err

        # Error if failed
        if not response.ok:
            error = f"Failed to update file '{file}' in db: {response.text}"
            LOG.exception(error)
            return updated_in_db, error

        try:
            updated_in_db, error = (True, response.json()["message"])
        except simplejson.JSONDecodeError as err:
            raise SystemExit from err

        return updated_in_db, error



# Standard library
import unittest.mock
import copy

# Installed
import json
import pytest
import requests

# Own modules
import dds_cli

RETURNED_PROJECTS_JSON = {
    "project_info": [
        {
            "Last updated": "Tue, 23 Nov 2021 10:27:42 GMT",
            "PI": "PI Name 1",
            "Project ID": "project_1",
            "Size": 20,
            "Status": "Available",
            "Title": "First Project",
        },
        {
            "Last updated": "Wed, 24 Nov 2021 10:27:42 GMT",
            "PI": "PI Name 2",
            "Project ID": "project_2",
            "Size": 30,
            "Status": "In Progress",
            "Title": "Second Project",
        },
    ],
    "total_size": 0,
    "total_usage": {"cost": 0.0, "usage": 0.0},
}


RETURNED_FILES_JSON = {
    "files_folders": [
        {"folder": False, "name": "simple_file.txt"},
        {"folder": False, "name": "simple_file2.txt"},
        {"folder": False, "name": "simple_file3.txt"},
        {"folder": True, "name": "subdir1"},
        {"folder": True, "name": "subdir3"},
        {"folder": True, "name": "subdir2"},
    ]
}

# Need to have two different ones since the dds code modifies the dictionary object
RETURNED_FILES_RECURSIVE_BOTTOM = {
    "files_folders": [
        {"folder": False, "name": "simple_file4.txt"},
        {"folder": False, "name": "simple_file5.txt"},
    ]
}


@pytest.fixture
def ls_runner(runner):
    """Run dds ls without a project specified."""

    def _run(cmd_list):
        return runner(cmd_list)

    yield _run


@pytest.fixture
def list_request():
    """A fixture that mocks the requests.get method.

    The functioned returned by this fixture takes parameters that adjust the status_code,
    return_json, ok, and side_effect.
    """
    with unittest.mock.patch.object(requests, "get") as mock_obj:

        def _request_mock(status_code, return_json=dict(), ok=True, side_effect=None):
            mock_returned_request = unittest.mock.MagicMock(status_code=status_code, ok=ok)
            if side_effect:
                mock_returned_request.json.side_effect = side_effect
            else:
                mock_returned_request.json.return_value = return_json
            mock_obj.return_value = mock_returned_request
            return mock_obj

        yield _request_mock


# ------------------------------------------------------------------------------------------------ #


def list_no_projects(ls_runner, list_request, command):
    """Perform test called by tests with different commands."""
    list_request_OK = list_request(200)
    result = ls_runner(command)

    assert result.exit_code == 0
    list_request_OK.assert_called_with(
        dds_cli.DDSEndpoint.LIST_PROJ,
        headers=unittest.mock.ANY,
        json={"usage": False},
    )

    assert "No project info was retrieved" in result.stderr
    assert "" == result.stdout


def test_list_no_projects_dds_ls(ls_runner, list_request):
    """Test that the list command works when no project is specified nor returned."""
    list_no_projects(ls_runner=ls_runner, list_request=list_request, command=["ls"])


def test_list_no_projects_dds_project_ls(ls_runner, list_request):
    """Test that the list command works when no project is specified nor returned."""
    list_no_projects(ls_runner=ls_runner, list_request=list_request, command=["project", "ls"])


# ------------------------------------------------------------------------------------------------ #


def list_no_project_specified(ls_runner, list_request, command):
    """Perform test called by tests with different commands."""
    list_request_OK = list_request(200, return_json=copy.deepcopy(RETURNED_PROJECTS_JSON))
    result = ls_runner(command)

    assert result.exit_code == 0
    list_request_OK.assert_called_with(
        dds_cli.DDSEndpoint.LIST_PROJ,
        headers=unittest.mock.ANY,
        json={"usage": False},
    )

    for substring in [
        "project_1",
        "project_2",
        "PI Name 1",
        "PI Name 2",
        "Available",
        "In Progress",
        "Tue, 23 Nov",
        "Wed, 24 Nov",
        "20 B",
        "30 B",
        "────────────────",  # Hack to test that there's a table printed
    ]:
        assert substring in result.stdout
    print(result.stderr)
    assert "" == result.stderr  # Click testing framework aborts any interactivity


def test_list_no_project_specified_dds_ls(ls_runner, list_request):
    """Test that the list command works when no project is specified."""
    list_no_project_specified(ls_runner=ls_runner, list_request=list_request, command=["ls"])


def test_list_no_project_specified_dds_project_ls(ls_runner, list_request):
    """Test that the list command works when no project is specified."""
    list_no_project_specified(
        ls_runner=ls_runner, list_request=list_request, command=["project", "ls"]
    )


# ------------------------------------------------------------------------------------------------ #


def list_no_project_specified_json(ls_runner, list_request, command):
    """Perform test called by tests with different commands."""
    list_request_OK = list_request(200, return_json=copy.deepcopy(RETURNED_PROJECTS_JSON))
    result = ls_runner(command)

    assert result.exit_code == 0
    list_request_OK.assert_called_with(
        dds_cli.DDSEndpoint.LIST_PROJ,
        headers=unittest.mock.ANY,
        json={"usage": False},
    )

    try:
        json_output = json.loads(result.stdout)
    except json.JSONDecodeError:
        assert False, "stdout is not JSON"

    project_ids = [project["Project ID"] for project in json_output]
    assert [
        "project_2",
        "project_1",
    ] == project_ids, "Default sorting of json should be last updated"


def test_list_no_project_specified_json_dds_ls(ls_runner, list_request):
    """Test that the list command works when no project is specified with json output."""
    list_no_project_specified_json(
        ls_runner=ls_runner, list_request=list_request, command=["ls", "--json"]
    )


def test_list_no_project_specified_json_dds_project_ls(ls_runner, list_request):
    """Test that the list command works when no project is specified with json output."""
    list_no_project_specified_json(
        ls_runner=ls_runner, list_request=list_request, command=["project", "ls", "--json"]
    )


# ------------------------------------------------------------------------------------------------ #


def list_no_project_specified_json_sort(ls_runner, list_request, command):
    """Perform test called by tests with different commands."""
    list_request_OK = list_request(200, return_json=copy.deepcopy(RETURNED_PROJECTS_JSON))
    result = ls_runner(command)

    assert result.exit_code == 0
    list_request_OK.assert_called_with(
        dds_cli.DDSEndpoint.LIST_PROJ,
        headers=unittest.mock.ANY,
        json={"usage": False},
    )

    try:
        json_output = json.loads(result.stdout)
    except json.JSONDecodeError:
        assert False, "stdout is not JSON"

    project_ids = [project["Project ID"] for project in json_output]
    assert ["project_1", "project_2"] == project_ids, "Sorting json on project id failed"


def test_list_no_project_specified_json_sort_dds_ls(ls_runner, list_request):
    """Test that the list command works when no project is specified with json output and non-default sorting."""
    list_no_project_specified_json_sort(
        ls_runner=ls_runner, list_request=list_request, command=["ls", "--json", "--sort", "id"]
    )


def test_list_no_project_specified_json_sort_dds_project_ls(ls_runner, list_request):
    """Test that the list command works when no project is specified with json output and non-default sorting."""
    list_no_project_specified_json_sort(
        ls_runner=ls_runner,
        list_request=list_request,
        command=["project", "ls", "--json", "--sort", "id"],
    )


# ------------------------------------------------------------------------------------------------ #


def list_with_project(ls_runner, list_request, command):
    """Perform test called by tests with different commands."""
    # Need to use deepcopy to be able to reuse the JSON object for other tests
    # since the DataLister.list_recursive uses pop on this dictionary
    list_request_OK = list_request(200, return_json=copy.deepcopy(RETURNED_FILES_JSON))
    result = ls_runner(command)

    assert result.exit_code == 0
    list_request_OK.assert_called_with(
        dds_cli.DDSEndpoint.LIST_FILES,
        params={"project": "project_1"},
        json={"subpath": None, "show_size": False},
        headers=unittest.mock.ANY,
    )
    print(result.stdout)
    for substring in [
        "Files / directories in project: project_1",
        "simple_file.txt",
        "simple_file2.txt",
        "simple_file3.txt",
        "subdir1",
        "subdir3",
        "subdir2",
    ]:
        assert substring in result.stdout


def test_list_with_project_dds_ls(ls_runner, list_request):
    """Test that the list command works when a project is specified."""
    list_with_project(
        ls_runner=ls_runner, list_request=list_request, command=["ls", "--project", "project_1"]
    )


def test_list_with_project_dds_data_ls(ls_runner, list_request):
    """Test that the list command works when a project is specified."""
    list_with_project(
        ls_runner=ls_runner,
        list_request=list_request,
        command=["data", "ls", "--project", "project_1"],
    )


# ------------------------------------------------------------------------------------------------ #


def list_with_project_and_tree(ls_runner, list_request, command):
    """Perform test called by tests with different commands."""
    list_request_OK = list_request(
        200,
        side_effect=[
            copy.deepcopy(RETURNED_FILES_JSON),
            copy.deepcopy(RETURNED_FILES_RECURSIVE_BOTTOM),
            copy.deepcopy(RETURNED_FILES_RECURSIVE_BOTTOM),
            copy.deepcopy(RETURNED_FILES_RECURSIVE_BOTTOM),
        ],
    )

    result = ls_runner(command)

    assert result.exit_code == 0
    list_request_OK.assert_any_call(
        dds_cli.DDSEndpoint.LIST_FILES,
        params={"project": "project_1"},
        json={"subpath": None, "show_size": False},
        headers=unittest.mock.ANY,
    )

    list_request_OK.assert_any_call(
        dds_cli.DDSEndpoint.LIST_FILES,
        params={"project": "project_1"},
        json={"subpath": "subdir1", "show_size": False},
        headers=unittest.mock.ANY,
    )

    list_request_OK.assert_any_call(
        dds_cli.DDSEndpoint.LIST_FILES,
        params={"project": "project_1"},
        json={"subpath": "subdir2", "show_size": False},
        headers=unittest.mock.ANY,
    )

    for substring in [
        "Files & directories in project: project_1",
        "simple_file.txt",
        "simple_file2.txt",
        "simple_file3.txt",
        "subdir1",
        "subdir2",
    ]:
        assert substring in result.stdout


def test_list_with_project_and_tree_dds_ls(ls_runner, list_request):
    """Test that the list command works when a project is specified."""
    list_with_project_and_tree(
        ls_runner=ls_runner,
        list_request=list_request,
        command=["ls", "--tree", "--project", "project_1"],
    )


def test_list_with_project_and_tree_dds_data_ls(ls_runner, list_request):
    """Test that the list command works when a project is specified."""
    list_with_project_and_tree(
        ls_runner=ls_runner,
        list_request=list_request,
        command=["data", "ls", "--tree", "--project", "project_1"],
    )


# ------------------------------------------------------------------------------------------------ #


def list_with_project_and_tree_json(ls_runner, list_request, command):
    """Perform test called by tests with different commands."""
    list_request_OK = list_request(
        200,
        side_effect=[
            copy.deepcopy(RETURNED_FILES_JSON),
            copy.deepcopy(RETURNED_FILES_RECURSIVE_BOTTOM),
            copy.deepcopy(RETURNED_FILES_RECURSIVE_BOTTOM),
            copy.deepcopy(RETURNED_FILES_RECURSIVE_BOTTOM),
        ],
    )

    result = ls_runner(command)

    assert result.exit_code == 0
    list_request_OK.assert_any_call(
        dds_cli.DDSEndpoint.LIST_FILES,
        params={"project": "project_1"},
        json={"subpath": None, "show_size": False},
        headers=unittest.mock.ANY,
    )

    list_request_OK.assert_any_call(
        dds_cli.DDSEndpoint.LIST_FILES,
        params={"project": "project_1"},
        json={"subpath": "subdir1", "show_size": False},
        headers=unittest.mock.ANY,
    )

    list_request_OK.assert_any_call(
        dds_cli.DDSEndpoint.LIST_FILES,
        params={"project": "project_1"},
        json={"subpath": "subdir2", "show_size": False},
        headers=unittest.mock.ANY,
    )

    try:
        json_output = json.loads(result.stdout)
    except json.JSONDecodeError:
        assert False, "stdout is not JSON"

    try:
        file = json_output["project_files_and_directories"]["subdir1"]["children"][
            "simple_file4.txt"
        ]
    except KeyError:
        assert False, f"wrong JSON structure: {json_output}"
    assert file["name"] == "simple_file4.txt"


def test_list_with_project_and_tree_json_dds_ls(ls_runner, list_request):
    """Test that the list command works when a project is specified."""
    list_with_project_and_tree_json(
        ls_runner=ls_runner,
        list_request=list_request,
        command=["ls", "--tree", "--json", "--project", "project_1"],
    )


def test_list_with_project_and_tree_json_dds_data_ls(ls_runner, list_request):
    """Test that the list command works when a project is specified."""
    list_with_project_and_tree_json(
        ls_runner=ls_runner,
        list_request=list_request,
        command=["data", "ls", "--tree", "--json", "--project", "project_1"],
    )


# ------------------------------------------------------------------------------------------------ #



"""Custom Exception classes."""

# Standard library
import logging
import requests

# Installed
import click

# Own modules

# Logger
LOG = logging.getLogger(__name__)


class InvalidMethodError(Exception):
    """Valid methods are only ls, put, get, rm. Anything else should raise errors."""

    def __init__(self, attempted_method, message="Attempting an invalid method in the DDS"):
        """Init invalid method error."""
        self.method = attempted_method
        self.message = message
        super().__init__(message)

    def __str__(self):
        """Print message and attempted method."""
        return f"{self.message}: {self.method}"


class DDSCLIException(click.ClickException):
    """Base exception for click in DDS."""

    def __init__(self, message, sign=":warning-emoji:"):
        """Init base exception."""
        self.message = message
        self.sign = sign
        super().__init__(message)

    def __str__(self):
        """Format error message and return with signs."""
        return f"{self.sign} {self.message} {self.sign}"


class AuthenticationError(click.ClickException):
    """Errors due to user authentication."""

    def __init__(self, message, sign=":no_entry_sign:"):
        """Update error message."""
        self.message = message
        self.sign = sign
        super().__init__(message)

    def __str__(self):
        """Return the message with Rich no-entry-sign emoji either side."""
        return f"{self.sign} {self.message} {self.sign}"


class TokenDeserializationError(Exception):
    """Error caused by being unable to deserialize the token."""

    def __init__(self, message):
        """Reformat error message."""
        super().__init__(message)


class TokenExpirationMissingError(Exception):
    """Error caused by missing token expiration time in the jose header of the token."""

    def __init__(self, message):
        """Reformat error message."""
        super().__init__(message)


class TokenNotFoundError(AuthenticationError):
    """No token retrieved from REST API or from File."""

    def __init__(self, message, sign=":warning-emoji:"):
        """Reformat error message."""
        super().__init__(message=message, sign=sign)


class ApiRequestError(requests.exceptions.RequestException):
    """Request to REST API failed."""

    def __init__(self, message):
        """Log and raise."""
        LOG.exception(message)
        super().__init__(message)


class ApiResponseError(Exception):
    """REST API Request does not return code 200 in response."""

    def __init__(self, message):
        """Log and raise."""
        LOG.exception(message)
        super().__init__(message)


class UploadError(Exception):
    """Errors relating to file uploads."""


class DownloadError(Exception):
    """Errors relating to file download."""


class NoDataError(Exception):
    """Errors when there is no data to do anything with."""


class APIError(Exception):
    """Error connecting to the dds web server."""



"""Data Remover -- Removes files from projects."""

###############################################################################
# IMPORTS ########################################################### IMPORTS #
###############################################################################

# Standard Library
import logging
import pathlib

# Installed
import requests
import rich
import rich.table
import rich.padding
import simplejson

# Own modules
import dds_cli
from dds_cli.custom_decorators import removal_spinner
from dds_cli import base
from dds_cli import DDSEndpoint

###############################################################################
# START LOGGING CONFIG ################################# START LOGGING CONFIG #
###############################################################################

LOG = logging.getLogger(__name__)

###############################################################################
# CLASSES ########################################################### CLASSES #
###############################################################################


class DataRemover(base.DDSBaseClass):
    """Data remover class."""

    def __init__(self, project: str, username: str, method: str = "rm", no_prompt: bool = False):
        """Handle actions regarding data deletion in the cli."""
        # Initiate DDSBaseClass to authenticate user
        super().__init__(username=username, project=project, method=method, no_prompt=no_prompt)

        self.failed_table = None
        self.failed_files = None

        # Only method "rm" can use the DataRemover class
        if self.method != "rm":
            raise dds_cli.exceptions.InvalidMethodError(
                attempted_method=method, message="DataRemover attempting unauthorized method"
            )

    def __create_failed_table(self, resp_json, level="File"):
        """Output a response after deletion."""
        # Check that enough info
        if not all(x in resp_json for x in ["not_exists", "not_removed"]):
            raise dds_cli.exceptions.APIError(
                f"Malformatted response detected when attempting remove action on {self.project}."
            )

        # Get info
        not_exists = resp_json["not_exists"]
        delete_failed = resp_json["not_removed"]

        # Create table if any files failed
        if not_exists or delete_failed:
            if self.no_prompt:
                self.failed_files = {"Errors": []}
                for x in not_exists:
                    self.failed_files["Errors"].append({x: f"No such {level.lower()}"})
                for x, y in delete_failed.items():
                    self.failed_files["Errors"].append({x: y})
            else:
                # Create table and add columns
                table = rich.table.Table(
                    title=f"{level}s not deleted",
                    title_justify="left",
                    show_header=True,
                    header_style="bold",
                )
                columns = [level, "Error"]
                for x in columns:
                    table.add_column(x)

                # Add rows
                for x in not_exists:
                    table.add_row(x, f"No such {level.lower()}")

                for x, y in delete_failed.items():
                    table.add_row(
                        f"[light_salmon3]{x}[/light_salmon3]",
                        f"[light_salmon3]{y}[/light_salmon3]",
                    )

                # Print out table
                self.failed_table = rich.padding.Padding(table, 1)

    @staticmethod
    def delete_tempfile(file: pathlib.Path):
        """Delete the specified file."""
        try:
            file.unlink()
        except FileNotFoundError as err:
            LOG.exception(str(err))
            LOG.warning("File deletion may have failed. Usage of space may increase.")

    # Public methods ###################### Public methods #
    @removal_spinner
    def remove_all(self, *_, **__):
        """Remove all files in project."""
        # Perform request to API to perform deletion
        try:
            response = requests.delete(
                DDSEndpoint.REMOVE_PROJ_CONT, params={"project": self.project}, headers=self.token
            )
        except requests.exceptions.RequestException as err:
            raise SystemExit from err

        if not response.ok:
            raise dds_cli.exceptions.APIError(f"Failed to delete files in project: {response.text}")

        # Print out response - deleted or not?
        try:
            resp_json = response.json()
        except simplejson.JSONDecodeError as err:
            raise SystemExit from err

        if "removed" not in resp_json:
            raise dds_cli.exceptions.APIError(
                "Malformatted response detected when attempting "
                f"to remove all files from {self.project}."
            )

    @removal_spinner
    def remove_file(self, files):
        """Remove specific files."""
        try:
            response = requests.delete(
                DDSEndpoint.REMOVE_FILE,
                params={"project": self.project},
                json=files,
                headers=self.token,
            )
        except requests.exceptions.RequestException as err:
            raise SystemExit from err

        if not response.ok:
            raise dds_cli.exceptions.APIError(
                f"Failed to delete file(s) '{files}' in project {self.project}: {response.text}"
            )

        # Get info in response
        try:
            resp_json = response.json()
        except simplejson.JSONDecodeError as err:
            raise SystemExit from err

        self.__create_failed_table(resp_json=resp_json)

    @removal_spinner
    def remove_folder(self, folder):
        """Remove specific folders."""
        try:
            response = requests.delete(
                DDSEndpoint.REMOVE_FOLDER,
                params={"project": self.project},
                json=folder,
                headers=self.token,
            )
        except requests.exceptions.RequestException as err:
            raise SystemExit from err

        if not response.ok:
            raise dds_cli.exceptions.APIError(
                f"Failed to delete folder(s) '{folder}' "
                f"in project {self.project}: {response.text}"
            )

        # Make sure required info is returned
        try:
            resp_json = response.json()
        except simplejson.JSONDecodeError as err:
            raise SystemExit from err

        self.__create_failed_table(resp_json=resp_json, level="Folder")
